/***************************************************************************
                          params.cpp  -  description
                             -------------------
    begin                : Sun May 8 2005
    copyright            : (C) 2005 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "params.h"
#include "common.h"
#include "externs.h"
#include "constants.h"
#include "string.h"
#include "s_specific_functions.h"
#include "s_hairpin_loop.h"
#include "s_stacked_pair.h"
#include "s_internal_loop.h"
#include "s_min_folding.h"
#include "s_sub_folding.h"
#include "s_energy_matrix.h"
#include "simfold.h"

// I played with the following functions to study a bit the parameters

void print_stacking_energies()
// prints the stacking energies
{
  int i,j,k,l;
  for (i=0; i < NUCL; i++)
      for (j=0; j < NUCL; j++)
          for (k=0; k < NUCL; k++)
          {
              for (l=0; l < NUCL; l++)
                  printf ("stack(%c,%c,%c,%c) = %d\n", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
              printf ("----\n");
          }
}



void print_tstacki_dangling_energies()
// prints the tstacki and dangling energies
{
    int met = 0;
    int i,j,k,l;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (tstacki[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("tstacki(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), tstacki[i][j][k][l]);
                        printf ("d3(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), dangle_top[i][j][k]);
                        printf ("d5(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(l), dangle_bot[i][j][l]);
                        printf ("Diff = %5d\n", tstacki[i][j][k][l] - dangle_top[i][j][k] - dangle_bot[i][j][l]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}


void print_stack_dangling_energies()
// prints the stack and dangling energies
{
    int met = 0;
    int i,j,k,l;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("stack(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
                        printf ("d3(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), dangle_top[i][j][k]);
                        printf ("d5(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(l), dangle_bot[i][j][l]);
                        printf ("Diff = %5d\n", stack[i][j][k][l] - dangle_top[i][j][k] - dangle_bot[i][j][l]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}


void print_stack_equation_dangling()
// prints the stack and dangling energies
{
    int met = 0;
    int i,j,k,l;
    float x,y,z,t;
    x = 0.0592;
    y = -7.0973;
    z = -0.4545;
    t = 7.7252;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("stack(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
                        printf (" eq = %5.2lf\n", x*dangle_top[i][j][k] + y*dangle_bot[i][j][l] + z*dangle_top[l][k][j] + t*dangle_bot[l][k][i]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}

void print_int22_tstacki()
// prints the int22 and tstacki energies
{
    int met = 0;
    int i,j,k,l,m,n,o,p;
    for (i=0; i < NUCL; i++)            
        for (j=0; j < NUCL; j++)    
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)            
                        for (n=0; n < NUCL; n++)    
                            for (o=0; o < NUCL; o++)
                                for (p=0; p < NUCL; p++)
                                {
                                    if (int22[i][j][k][l][m][n][o][p] < INF)
                                    {
                                        printf ("int22(%c,%c,%c,%c,%c,%c,%c,%c) = %5d -- %5d = \t", int_to_nuc(i), int_to_nuc(j), 
                                            int_to_nuc(k), int_to_nuc(l), int_to_nuc(m), int_to_nuc(n), int_to_nuc(o), int_to_nuc(p),
                                            int22[i][j][k][l][m][n][o][p], tstacki[i][j][k][l]+tstacki[n][m][p][o]);
                                        printf ("tstacki(%c,%c,%c,%c) + tstacki(%c,%c,%c,%c) = %5d + %5d\n",    
                                            int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), 
                                            int_to_nuc(n), int_to_nuc(m), int_to_nuc(p), int_to_nuc(o), tstacki[i][j][k][l], tstacki[n][m][p][o]);
                                    }
                                }
}


void print_int22()
// prints the int22 and tstacki energies
{
    int met = 0;
    int i,j,k,l,m,n,o,p;
    for (i=0; i < NUCL; i++)            
        for (j=0; j < NUCL; j++)
            if (can_pair(i,j))
            { 
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)            
                            for (n=0; n < NUCL; n++)
                                if (can_pair(m,n))
                                {
                                    for (o=0; o < NUCL; o++)
                                        for (p=0; p < NUCL; p++)
                                        {
                                            // exclude duplicates
                                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                            {    
                                                printf ("%.2lf\n",
                                                //int22[%c][%c][%c][%c][%c][%c][%c][%c]=
                                                //int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l),
                                                //int_to_nuc(m), int_to_nuc(n), int_to_nuc(o), int_to_nuc(p),
                                                int22[i][j][k][l][m][n][o][p]/100.0);
                                            }
                                        }
                                }                                        
            }
}


void test_int22_differences()
// there are some differences here, because in the 53-param model I didn't include all the measured internal loops
{
    int sint  [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL];
    int i,j,k,l,m,n,o,p;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                            for (o=0; o < NUCL; o++)
                                for (p=0; p < NUCL; p++)
                                {
                                    sint[i][j][k][l][m][n][o][p] = int22[i][j][k][l][m][n][o][p];
                                    int22[i][j][k][l][m][n][o][p] = INF;
                                }


    simple_internal_energy = 0;
    //save_parameters("turner_parameters_fmnew.txt");
    num_params = create_string_params();

    fill_data_structures_with_new_parameters("turner_parameters_fmnew.txt");

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                            for (o=0; o < NUCL; o++)
                                for (p=0; p < NUCL; p++)
                                {
                                    if (sint[i][j][k][l][m][n][o][p] != int22[i][j][k][l][m][n][o][p])
                                        //printf ("int22[%d][%d][%d][%d][%d][%d][%d][%d]  tur=%d, com=%d, diff=%d\n", 
                                        //          i,j,k,l,m,n,o,p, sint[i][j][k][l][m][n][o][p], int22[i][j][k][l][m][n][o][p], int22[i][j][k][l][m][n][o][p]-sint[i][j][k][l][m][n][o][p]);
                                        printf ("int22[5'%c%c%c%c/3'%c%c%c%c]    T99=%3d    MAextra=%3d    T99-MAextra=%d\n",
                                                  int_to_nuc(i), int_to_nuc(k), int_to_nuc(o), int_to_nuc(m), int_to_nuc(j), int_to_nuc(l), int_to_nuc(p), int_to_nuc(n), sint[i][j][k][l][m][n][o][p], int22[i][j][k][l][m][n][o][p], sint[i][j][k][l][m][n][o][p]-int22[i][j][k][l][m][n][o][p]);
                                } 
}

void test_tstacki_differences()
// this is fine, no differences here
{
    int sint  [NUCL] [NUCL] [NUCL] [NUCL];
    int i,j,k,l;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    sint[i][j][k][l] = tstacki[i][j][k][l];
                }


    simple_internal_energy = 0;
    //save_parameters("turner_parameters_fmnew.txt");
    num_params = create_string_params();

    fill_data_structures_with_new_parameters("turner_parameters_fmnew.txt");

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (sint[i][j][k][l] != tstacki[i][j][k][l])
                        printf ("tstacki[%d][%d][%d][%d]  tur=%d, com=%d, diff=%d\n",
                                    i,j,k,l, sint[i][j][k][l], tstacki[i][j][k][l], sint[i][j][k][l]-tstacki[i][j][k][l]);
                }
}


void test_int21_differences()
// they are the same as turner's, within 0.05 kcal/mol
{
    int sint  [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL];
    int i,j,k,l,m,n,o;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                            for (o=0; o < NUCL; o++)
                            {
                                sint[i][j][k][l][m][n][o] = int21[i][j][k][l][m][n][o];
                            }

    fill_data_structures_with_new_parameters("turner_parameters_fmnew.txt");

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                            for (o=0; o < NUCL; o++)
                            {
                                if (sint[i][j][k][l][m][n][o] != int21[i][j][k][l][m][n][o])
                                    printf ("int21[%d][%d][%d][%d][%d][%d][%d]  tur=%d, com=%d, diff=%d\n",
                                                i,j,k,l,m,n,o, sint[i][j][k][l][m][n][o], int21[i][j][k][l][m][n][o], int21[i][j][k][l][m][n][o]-sint[i][j][k][l][m][n][o]);
                            }
}


void test_int11_differences()
// they are the same as turner's, within 0.05 kcal/mol
{
    int sint  [NUCL] [NUCL] [NUCL] [NUCL] [NUCL] [NUCL];
    int i,j,k,l,m,n,o;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                        {
                            sint[i][j][k][l][m][n] = int11[i][j][k][l][m][n];
                            int11[i][j][k][l][m][n] = INF;
                        }

    fill_data_structures_with_new_parameters("turner_parameters_fmnew.txt");

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                        {
                            if (sint[i][j][k][l][m][n] != int11[i][j][k][l][m][n])
                                printf ("int11[%d][%d][%d][%d][%d][%d]  tur=%d, com=%d, diff=%d\n",
                                            i,j,k,l,m,n, sint[i][j][k][l][m][n], int11[i][j][k][l][m][n], int11[i][j][k][l][m][n]-sint[i][j][k][l][m][n]);
                        }
}

/////////////////////////////////////////////////////////////
// next functions are for playing with parameter learning. //
/////////////////////////////////////////////////////////////


int structure_type_index (char type[])
  // Mirela: Nov 23, 2003
  // Given the type as a string, return the index in string_params
{
  int i, found;
  found = 0;
  
  // if type is internal/hairpin/bulge penalty by size, which is bigger than MAXLOOP_X, 
  //    then type is actually MAXLOOP_X
  int size;
  // do it later, only if this is penalty, for faster runtime. This actually helps a bit.
 

  // just a hack to make this faster than linear. NOT GOOD IF THE (NUMBER OF) PARAMETERS CHANGES!!!
    int start = 0;
    if (type[0] == 's')     // stack
        start = 0;
    else if (type[0] == 't')     // tstack or tloop
    {
        if (type[1] == 's')  // tstack
        {
            if (type[8] == '0')   // tstackh[0]
                start = 21;
            else if (type[8] == '1')    //tstackh[1]
                start = 37;
            else if (type[8] == '2')    //tstackh[2]
                start = 53;
            else if (type[8] == '3')    //tstackh[3]
                start = 85;
        }
        else if (type[1] == 'l')    // tloop
            start = 333;
    }    
    else if (type[0] == 'm')    // misc....
    {
        if (type[5] == 'i')
        {
            if (type[13] == '_')    // misc.internal_
                start = 117;
            else if (type[13] == '1')   // misc.internal11_
                start = 151;
            else if (type[13] == '2')
            {
                if (type[15] == '1')     // misc.internal21_
                    start = 205;
                else if (type[15] == '2')   // misc.internal22_
                    start = 255;
            }            
        }
        else    // al other misc
            start = 324;
    }
    else if (type[0] == 'i')    // int11, int21, int22 or internal_penalty_by_size
    {
        if (type[3] == '1')  // int11
            start = 120;
        else if (type[3] == '2')    // int21 or int22
        {
            if (type[4] == '1')  // int21
            {
                if (type[6] == '1')     // int21[1]...
                    start = 153;
                else if (type[6] == '2')     // int21[2]...
                    start = 179;
            }
            else if (type[4] == '2')  // int22
            {
                if (type[6] == '0')     // int22[0]...
                    start = 207;
                else if (type[6] == '1')     // int22[1]...
                    start = 219;
                else if (type[6] == '2')     // int22[2]...
                    start = 231;
                else if (type[6] == '3')     // int22[3]...
                    start = 243;
            }
        }
        else if (type[3] == 'e')    // internal_penalty_by_size
        {
            start = 308;
            // next internal
            if (sscanf (type, "internal_penalty_by_size[%d]", &size))
            {
                if (size > MAXLOOP_I)
                {
                    //printf ("size = %d\n", size);
                    sprintf (type, "internal_penalty_by_size[%d]", MAXLOOP_I);
                }
            }            
        }
    }
    else if (type[0] == 'd')    // dangle_top or dangle_bot
    {
        if (type[7] == 't')  // dangle_top
            start = 260;
        else if (type[7] == 'b')  // dangle_bot
            start = 284;
    }
    else if (type[0] == 'b')    // bulge_penalty_by_size
    {
        start = 311;
        // next bulge
        if (sscanf (type, "bulge_penalty_by_size[%d]", &size))
        {
            if (size > MAXLOOP_B)
            {
                //printf ("size = %d\n", size);
                sprintf (type, "bulge_penalty_by_size[%d]", MAXLOOP_B);
            }
        }        
    }
    else if (type[0] == 'h')    // hairpin_penalty_by_size
    {
        start = 317;
        // first hairpin
        if (sscanf (type, "hairpin_penalty_by_size[%d]", &size))
        {
            if (size > MAXLOOP_H)
            {
                //printf ("size = %d\n", size);
                sprintf (type, "hairpin_penalty_by_size[%d]", MAXLOOP_H);
            }
        }        
    }


  
  // now traverse all params, to see which is the right index
  // veeery slow!!! // it's faster now, since I added the above 
  for (i=start; i < num_params; i++)
    {
      if (strcmp (type, string_params[i]) == 0)
        // Note: this is very slow
        {
          found = 1;
          //printf ("%s found in %d steps\n", type, i-start+1);
          break;
        }
    }
  if (!found)
    {
      printf ("!!!TYPE NOT found: %s!!!\n", type);
      exit(1);
    }
  return i;
}



void count_AU_penalty (int base_i, int base_j, double *counter)
// PRE:  base_i and base_j make a pair
// POST: increment counter at the appropriate position
// Mirela: Nov 23, 2003
{
  int index;
    if ((base_i != C || base_j != G) &&
        (base_i != G || base_j != C))      
      {
        index = structure_type_index ("misc.terminal_AU_penalty");
        counter[index]++;
      }
}


void count_dangling_energy (int *sequence, int link, int i1, int i2, int i3, int i4, double *counter)
//      (   )...(   )
//      i1..i1..i3..i4
// PRE:  (i1, i2) and (i3, i4) are pairs, i2 and i3 are neighbours, i2 < i3
// POST: return dangling energy between i2 and i3
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;
    int first_index;    // first index should be dangle_top[0][3][0]
    first_index = structure_type_index ("dangle_top[0][3][0]");
    
    d_top = 0;
    d_bot = 0;

    if (i2 != link)
      {
        //d_top = MIN (0, IGINF(dangle_top[sequence[i2]][sequence[i1]][sequence[i2+1]]));
        d_top = dangle_top[sequence[i2]] [sequence[i1]] [sequence[i2+1]];
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i2], sequence[i1], sequence[i2+1]);
        index_top = structure_type_index (type);
      }
    if (i3-1 != link)
      {
        //d_bot = MIN (0, IGINF(dangle_bot[sequence[i4]] [sequence[i3]] [sequence[i3-1]]));
        d_bot = dangle_bot[sequence[i4]] [sequence[i3]] [sequence[i3-1]];
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i3-1]);
        index_bot = structure_type_index (type);
      }
    if (i2+1 == i3-1 && i2 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i2+1 == i3-1 && i3-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i2+1 == i3-1)     // see which is smaller
    {
        //energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        // NO! This is not right if we don't know which of d_top and d_bot is smaller        

        // if we restrict the 3' dangling ends to be less than the 5' ones, then it's ok to do what follows

        // if we fix the dangling ends to the Turner parameters, we have to count this
        //if (d_top < d_bot) counter[index_top]++;
        //else counter[index_bot]++;

        if (simple_dangling_ends)
        {
            energy = d_top;
            counter[index_top]++;
        }
        else
        {
            if (d_top < d_bot)
            {
                energy = d_top;
                counter[index_top]++;
            }
            else
            {
                energy = d_bot;
                counter[index_bot]++;
            }
        }

        // if we introduce another variable as min, we need to do this
        //counter_min_dangle[index_top-first_index][index_bot-first_index]++;
    }
    else if (i2+1 < i3-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }

    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}

void count_dangling_energy_left (int *sequence, int link, int i1, int i2, int i3, int i4, double *counter)
//      (....(    )   )
//      i1   i3  i4  i2
// PRE:  (i1, i2) and (i3, i4) are pairs, i1 and i3 are neighbours, i3 < i2
// POST: return dangling energy between i1 and i3
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;
    d_top = 0;
    d_bot = 0;
    int first_index;    // first index should be dangle_top[0][3][0]
    first_index = structure_type_index ("dangle_top[0][3][0]");

    // this will be used in multi-loops.
    // add the dangle_top, even if it is positive
    if (i1 != link)
      {
        //d_top = MIN (0, IGINF(dangle_top[sequence[i1]] [sequence[i2]] [sequence[i1+1]]));
        d_top = dangle_top[sequence[i1]] [sequence[i2]] [sequence[i1+1]];
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i1], sequence[i2], sequence[i1+1]);
        index_top = structure_type_index (type);
      }
    // in the other parts of the multi-loop, the dangles are added only if they are negative
    if (i3-1 != link)
      {
        //d_bot = MIN (0, IGINF(dangle_bot[sequence[i4]] [sequence[i3]] [sequence[i3-1]]));
        d_bot = dangle_bot[sequence[i4]] [sequence[i3]] [sequence[i3-1]];
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i3-1]);
        index_bot = structure_type_index (type);
      }
    if (i1+1 == i3-1 && i1 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i1+1 == i3-1 && i3-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i1+1 == i3-1)     // see which is smaller
    {
        //energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        //if (d_top < d_bot) counter[index_top]++;
        //else counter[index_bot]++;
        //counter_min_dangle[index_top-first_index][index_bot-first_index]++;

        if (simple_dangling_ends)
        {
            energy = d_top;
            counter[index_top]++;
        }
        else
        {
            if (d_top < d_bot)
            {
                energy = d_top;
                counter[index_top]++;
            }
            else
            {
                energy = d_bot;
                counter[index_bot]++;
            }
        }
    }
    else if (i1+1 < i3-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }
    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}


void count_dangling_energy_right (int *sequence, int link, int i1, int i2, int i3, int i4, double *counter)
//      (    (    )...)
//      i1   i3  i4  i2
// PRE:  (i1, i2) and (i3, i4) are pairs, i1 and i3 are neighbours, i3 < i2
// POST: return dangling energy between i4 and i2
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;
    int first_index;    // first index should be dangle_top[0][3][0]
    first_index = structure_type_index ("dangle_top[0][3][0]");
    
    d_top = 0;
    d_bot = 0;

    if (i4 != link)
      {
        //d_top = MIN (0, IGINF(dangle_top[sequence[i4]] [sequence[i3]] [sequence[i4+1]]));
        d_top = dangle_top[sequence[i4]] [sequence[i3]] [sequence[i4+1]];
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i4+1]);
        index_top = structure_type_index (type);
      }
    if (i2-1 != link)
      {
        //d_bot = MIN (0, IGINF(dangle_bot[sequence[i1]] [sequence[i2]] [sequence[i2-1]]));
        d_bot = dangle_bot[sequence[i1]] [sequence[i2]] [sequence[i2-1]];
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i1], sequence[i2], sequence[i2-1]);
        index_bot = structure_type_index (type);
      }
    if (i4+1 == i2-1 && i4 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i4+1 == i2-1 && i2-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i4+1 == i2-1)     // see which is smaller
    {
        //energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        //if (d_top < d_bot) counter[index_top]++;
        //else counter[index_bot]++;
        //counter_min_dangle[index_top-first_index][index_bot-first_index]++;
        
        if (simple_dangling_ends)
        {
            energy = d_top;
            counter[index_top]++;
        }
        else
        {
            if (d_top < d_bot)
            {
                energy = d_top;
                counter[index_top]++;
            }
            else
            {
                energy = d_bot;
                counter[index_bot]++;
            }
        }
    }
    else if (i4+1 < i2-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }
    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}



void count_penalty_by_size (int size, char type, double *counter)
// PRE:  size is the size of the loop
//       type is HAIRP or INTER or BULGE
// POST: return the penalty by size of the loop
// Mirela: Nov 23, 2003
{
    char t[100];
    int index;
    int penaltyMAX, penalty;
    double logval;

    // the penalties for size <= MAXLOOP should be read from the file "loop"
    // actually only up to some size are measured, for the others apply the formula (see constants.h)
    if (type == 'H' && size <= MAXLOOP_H)
    {
        //return hairpin_penalty_by_size[size];
        sprintf (t, "hairpin_penalty_by_size[%d]", size);
        index = structure_type_index(t);
        counter[index]++;
        return;
    }
    if (type == 'I' && size <= MAXLOOP_I)
    {
        //return internal_penalty_by_size[size];
        sprintf (t, "internal_penalty_by_size[%d]", size);
        index = structure_type_index(t);
        counter[index]++;
        return;
    }
    if (type == 'B' && size <= MAXLOOP_B)
    {          
        //return bulge_penalty_by_size[size];
        sprintf (t, "bulge_penalty_by_size[%d]", size);
        index = structure_type_index(t);
        counter[index]++;
        return;        
    }

    // Mirela: Nov 23, 2003
    // We keep the formula 
    // Extrapolation for large loops based on polymer theory 
    // internal, bulge or hairpin loops > 30: dS(T)=dS(30)+param*ln(n/30) 
    // and imcrement for x_penalty_by_size[MAXLOOP] and for misc.param_greater30

    // size > MAXLOOP _B, _H or _I
    if (type == 'H')
      {
        penaltyMAX = hairpin_penalty_by_size[MAXLOOP_H];
        logval = log (1.0*size/MAXLOOP_H);
        sprintf (t, "hairpin_penalty_by_size[%d]", MAXLOOP_H);
        index = structure_type_index (t);
        counter[index]++;
        // added next line on Dec 23, 2006        
        counter[num_params] += round(1.079*logval*100.0);
      }
    else if (type == 'I')
      {
        penaltyMAX = internal_penalty_by_size[MAXLOOP_I];
        logval = log (1.0*size/MAXLOOP_I);
        sprintf (t, "internal_penalty_by_size[%d]", MAXLOOP_I);
        index = structure_type_index (t);
        counter[index]++;
        // added next line on Dec 23, 2006
        //printf ("Adding %d to total\n", (int)(1.079*logval*100.0));
        counter[num_params] += round(1.079*logval*100.0);        
      }
    else
      {
        penaltyMAX = bulge_penalty_by_size[MAXLOOP_B];
        logval = log (1.0*size/MAXLOOP_B);
        sprintf (t, "bulge_penalty_by_size[%d]", MAXLOOP_B);
        index = structure_type_index (t);
        counter[index]++;
        // added next line on Dec 23, 2006
        counter[num_params] += round(1.079*logval*100.0);
      }
    
    // Dec 4, 2005: let's keep misc.param_greater30 fixed, i.e. not consider it a parameter
    penalty = (int) (penaltyMAX + misc.param_greater30 * logval);
    //index = structure_type_index ("misc.param_greater30");
    // TODO: to store double instead of int
    //counter[index]+= (int) LogValue[size];

    //    return penalty;
    return;
}

void count_asymmetry_penalty (int size1, int size2, double *counter)
// PRE:  size1 and size2 are the sizes of the two free base groups
// POST: Calculate the asymmetry penalty for internal loops
//       Note that if size1 == size2, pen is 0
// Mirela: Nov 23, 2003
{
    // count a fixed value for now
    int ass, pen, index;    
    ass = misc.asymmetry_penalty_array [MIN (2, MIN (size1, size2))-1];
    pen = MIN (misc.asymmetry_penalty_max_correction, abs (size1-size2) * ass);

    //printf ("Adding asym penalty %d to total\n", pen);
    counter[num_params] += pen;
    
    // don't count anything for now, considered them fixed. 
    // TODO: to add new parameter t, for the min(a,b), and then add the restriction t <= a, t <= b and minimize (-t)
    /*
    int ass, pen, index;    
    ass = misc.asymmetry_penalty_array [MIN (2, MIN (size1, size2))-1];
    pen = MIN (misc.asymmetry_penalty_max_correction, abs (size1-size2) * ass);
    
    // store misc.asymmetry_penalty_array[0] and [1], and misc.asymmetry_penalty_max_correction
    index = structure_type_index("misc.asymmetry_penalty_max_correction"); counter[index]++;
    if (size1-size2 == 1)
    {
        index = structure_type_index("misc.asymmetry_penalty_array[0]"); counter[index]++;
    }
    else
    {
        index = structure_type_index("misc.asymmetry_penalty_array[1]"); counter[index]++;
        // [2] and [3] are never used
    }
    */
    // pen can only be 0.5, 1, 1.5, 2, 2.5 or 3 with the current model;
    /*switch (size1-size2)
      {
        case 1: index = structure_type_index("misc.asymmetry_penalty[1]"); counter[index]++; break;
        case 2: index = structure_type_index("misc.asymmetry_penalty[2]"); counter[index]++; break;
        case 3: index = structure_type_index("misc.asymmetry_penalty[3]"); counter[index]++; break;
        case 4: index = structure_type_index("misc.asymmetry_penalty[4]"); counter[index]++; break;
        case 5: index = structure_type_index("misc.asymmetry_penalty[5]"); counter[index]++; break;
        default:index = structure_type_index("misc.asymmetry_penalty[6]"); counter[index]++; break;          
      }*/
    
    //    return pen;
}


void count_types (int link, int *sequence, char *csequence, str_features *f, double *counter)
// Mirela: Nov 23, 2003
// PRE: string_params have been filled, i.e. by num_params = calling create_string_params() or num_params = create_building_block_strings()
//        where num_params is the global variable
// instead of summing up the energies of each type, increment the corresponding counter
{
    int energy, i, en, AUpen;
    int dang;
    int misc_energy;
    int h,l;    
    static int cannot_add_dangling[MAXSLEN];
    char type[100];
    int index;
    
    int nb_nucleotides = strlen(csequence);
    for (i=0; i < nb_nucleotides; i++) cannot_add_dangling[i] = 0;

    energy = 0;
    AUpen = 0;       
    
    for (i=0; i < nb_nucleotides; i++)
    {
    
        // add some AU_penalties
        if ((i==0 || (i-1 == link && !cannot_add_dangling[i-1] && f[i-1].pair == -1))
             && f[i].pair > i)
        {
            AUpen = AU_penalty (sequence[i], sequence[f[i].pair]);
            if (debug)
                printf ("%d - AUpen1 \t- add energy %6d\n", i, AUpen);            
            energy += AUpen;            
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
        }            
        else if ( i > 0 && f[i].pair > i && f[i-1].pair < i-1 &&
             f[i-1].pair != -1 && !cannot_add_dangling[i])
            //  )(  
        {            
            AUpen = AU_penalty (sequence[i], sequence[f[i].pair]);
            if (debug)
                printf ("%d - AUpen2 \t- add energy %6d\n", i, AUpen);
            energy += AUpen; 
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
        }            
    
        // add dangling energies and AU_penalties
        if (f[i].pair == -1 && !cannot_add_dangling[i])
        {
            if ((i == 0 || i-1 == link || (i > 0 && f[i-1].pair == -1 && i != link)) &&
                 i < nb_nucleotides-1 && f[i+1].pair > i+1)
                // .( or ..(
            {
                if (no_dangling_ends)
                    dang = 0;
                else
                {                                
                    //dang = MIN (0, IGINF(dangle_bot [sequence[f[i+1].pair]] [sequence[i+1]] [sequence[i]]));
                    dang = IGINF(dangle_bot [sequence[f[i+1].pair]] [sequence[i+1]] [sequence[i]]);
                    sprintf (type, "dangle_bot[%d][%d][%d]",sequence[f[i+1].pair], sequence[i+1], sequence[i]);
                    index = structure_type_index (type);
                    counter[index]++;                    
                }
                AUpen = AU_penalty (sequence[i+1], sequence[f[i+1].pair]);
                if (debug)
                {
                    printf ("%d - dangle1 \t- add energy %6d\n", i, dang);
                    printf ("%d - AUpen3 \t- add energy %6d\n", i, AUpen);
                }                    
                energy += dang + AUpen;
                count_AU_penalty (sequence[i+1], sequence[f[i+1].pair], counter);
            }
            else if ((i == nb_nucleotides-1 || i == link ||
                     (i < nb_nucleotides-1 && f[i+1].pair == -1 && i-1 != link)) &&
                     i > 0 && f[i-1].pair > -1 && f[i-1].pair < i-1)
                // ). or )..
            {
                if (no_dangling_ends)
                    dang = 0;
                else
                {                                            
                    //dang = MIN (0, IGINF(dangle_top [sequence[i-1]] [sequence[f[i-1].pair]] [sequence[i]]));
                    dang = IGINF(dangle_top [sequence[i-1]] [sequence[f[i-1].pair]] [sequence[i]]);
                    if (debug)                
                        printf ("%d - dangle2 \t- add energy %6d\n", i, dang);
                    energy += dang;
                    sprintf (type, "dangle_top[%d][%d][%d]",sequence[i-1], sequence[f[i-1].pair], sequence[i]);
                    index = structure_type_index (type);
                    counter[index]++;
                }
            }
            else if (i < nb_nucleotides-1 && f[i+1].pair > i+1 && f[i-1].pair < i-1 && f[i-1].pair != -1)
               // ).( 
            {
                if (no_dangling_ends)
                    dang = 0;
                else
                {                                            
                    //dang = MIN (0, IGINF(s_dangling_energy (sequence, f[i-1].pair, i-1, i+1, f[i+1].pair)));
                    dang = IGINF(s_dangling_energy (sequence, f[i-1].pair, i-1, i+1, f[i+1].pair));
                    count_dangling_energy (sequence, link, f[i-1].pair, i-1, i+1, f[i+1].pair, counter);
                }
                AUpen = AU_penalty (sequence[i+1], sequence[f[i+1].pair]);
                if (debug)
                {              
                    printf ("%d - dangle1 \t- add energy %6d\n", i, dang);
                    printf ("%d - AUpen4 \t- add energy %6d\n", i, AUpen);
                }    
                energy += dang + AUpen;                
                count_AU_penalty (sequence[i+1], sequence[f[i+1].pair], counter);
            }
            else
            {
                continue;
            }
        }
        
        if (f[i].pair < i)
        {
            continue;
        }       

        if (f[i].type == STACK)
        {
            en = s_stacked_pair::get_energy (i, f[i].pair, sequence);
            if (debug)            
                printf ("%d stack \t- add energy %6d\n", i, en);        
            energy += en;
            s_stacked_pair::count_get_energy (i, f[i].pair, sequence, counter);
        }
        else if (f[i].type == HAIRP)
        {
            if (link > -1 && i <= link && link < f[i].pair)
            {    
                // add intermolecular initiation
                misc_energy = misc.intermolecular_initiation;
                index = structure_type_index ("misc.intermolecular_initiation");
                if (debug)
                    printf ("%d intermol \t- add energy %6d\n", i, misc.intermolecular_initiation);                
                counter[index]++;
                energy += misc_energy;
                // add AU penalty
                count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
                AUpen = AU_penalty (sequence[i], sequence[f[i].pair]);
                energy += AUpen;
                // add dangling ends
                if (link > i)    // there is a dangle_top    (. )
                {                
                    //dang = MIN (0, IGINF(dangle_top [sequence[i]] [sequence[f[i].pair]] [sequence[i+1]]));
                    dang = IGINF(dangle_top [sequence[i]] [sequence[f[i].pair]] [sequence[i+1]]);
                    if (debug)
                    {
                        printf ("%d - dangle-spec-hairp \t- add energy %6d\n", i, dang);
                    }                    
                    energy += dang;
                    sprintf (type, "dangle_top[%d][%d][%d]", sequence[i], sequence[f[i].pair], sequence[i+1]);
                    index = structure_type_index (type);
                    counter[index]++;
                }                 
                if (link < f[i].pair-1)    // there is a dangle_bot    ( .)
                {                
                    //dang = MIN (0, IGINF(dangle_bot [sequence[i]] [sequence[f[i].pair]] [sequence[f[i].pair-1]]));
                    dang = IGINF(dangle_bot [sequence[i]] [sequence[f[i].pair]] [sequence[f[i].pair-1]]);
                    if (debug)
                    {
                        printf ("%d - dangle-spec-hairp \t- add energy %6d\n", i, dang);
                    }                    
                    energy += dang;
                    sprintf (type, "dangle_bot[%d][%d][%d]", sequence[i], sequence[f[i].pair], sequence[f[i].pair-1]);
                    index = structure_type_index (type);
                    counter[index]++;
                }                 
                
            }
            else
            {                
                en = s_hairpin_loop::get_energy (i, f[i].pair, sequence, csequence);
                if (debug)
                    printf ("%d hairpin \t- add energy %6d\n", i, en);
                energy += en;
                s_hairpin_loop::count_get_energy (i, f[i].pair, sequence, csequence, counter);
            }    
        }
        else if (f[i].type == INTER)
        {
            int ip, jp;
            ip = f[i].bri[0];
            jp = f[f[i].bri[0]].pair;
            cannot_add_dangling[ip-1] = 1;
            cannot_add_dangling[jp+1] = 1;
            en = s_internal_loop::get_energy (i, f[i].pair, ip, jp, sequence);
            if (debug)
                printf ("%d internal \t- add energy %6d\n", i, en);        
            energy += en;
            s_internal_loop::count_get_energy (i, f[i].pair, ip, jp, sequence, counter);
        }
        else  // (f[i].type == MULTI)
        {
            dang = 0;
            misc_energy = 0;
            AUpen = 0;
            int special;
            special = 0;
            // add the energies/enthalpies for free bases
            // first find out if it is a regular multi-loop or a special multi-loop
            l = i;

            if (f[i].num_branches > MAX_BRANCHES)
            {
                printf ("ERROR! f[i].num_branches=%d, but MAX_BRANCHES=%d\n", f[i].num_branches, MAX_BRANCHES);
                exit(1);
            }
            while (l < f[i].bri[0] && !special)
                if (l++ == link) special = 1;
            h = 0;                
            while (h < f[i].num_branches-1 && !special)
            {
                l = f[f[i].bri[h]].pair;
                while (l < f[i].bri[h+1] && !special)
                    if (l++ == link) special = 1;
                h++;    
            }
            l = f[f[i].bri[f[i].num_branches-1]].pair;
            while (l < f[i].pair && !special)
                if (l++ == link) special = 1;
            // now we now if it a special multi-loop or not                                  

            if (!special)
            {
//                printf ("Regular ML\n");
              index = structure_type_index ("misc.multi_free_base_penalty");
                for (l=i+1; l < f[i].bri[0]; l++)
                  {
                    misc_energy += misc.multi_free_base_penalty;
                    counter[index]++;
                  }
                for (h=0; h < f[i].num_branches-1; h++)
                {
                    for (l = f[f[i].bri[h]].pair + 1; l < f[i].bri[h+1]; l++)
                      {
                        misc_energy += misc.multi_free_base_penalty;
                        counter[index]++;
                      }
                }
                for (l = f[f[i].bri[f[i].num_branches-1]].pair + 1; l < f[i].pair; l++)
                  {
                    misc_energy += misc.multi_free_base_penalty;
                    counter[index]++;
                  }
                misc_energy += misc.multi_offset;
                misc_energy += misc.multi_helix_penalty * (f[i].num_branches + 1);
                index = structure_type_index ("misc.multi_offset");
                counter[index]++;
                index = structure_type_index ("misc.multi_helix_penalty");
                counter[index]+= f[i].num_branches + 1;
            }
            /*
            else
            {
//                printf ("Special ML\n");            
                misc_energy = misc.intermolecular_initiation;
                index = structure_type_index ("misc.intermolecular_initiation");
                counter[index]++;
            } 
            */               
            // add AU_penalties for multi-loop
            AUpen += AU_penalty (sequence[i], sequence[f[i].pair]);
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
            for (h=0; h < f[i].num_branches; h++)
              {
                AUpen += AU_penalty (sequence[f[i].bri[h]],sequence[f[f[i].bri[h]].pair]);
                count_AU_penalty (sequence[f[i].bri[h]],sequence[f[f[i].bri[h]].pair], counter);
              }
        
            // add dangling energies for multi-loop
            if (no_dangling_ends)
                dang = 0;
            else
            {                                            
                dang += s_dangling_energy_left (sequence, i, f[i].pair, f[i].bri[0], f[f[i].bri[0]].pair);
                count_dangling_energy_left (sequence, link, i, f[i].pair, f[i].bri[0], f[f[i].bri[0]].pair, counter);
                for (l=0; l < f[i].num_branches - 1; l++)
                {
                    dang += s_dangling_energy (sequence, f[i].bri[l], f[f[i].bri[l]].pair, f[i].bri[l+1], f[f[i].bri[l+1]].pair);
                    count_dangling_energy (sequence, link, f[i].bri[l], f[f[i].bri[l]].pair, f[i].bri[l+1], f[f[i].bri[l+1]].pair, counter);
                }
                dang += s_dangling_energy_right (sequence, i, f[i].pair, f[i].bri[f[i].num_branches-1], f[f[i].bri[f[i].num_branches-1]].pair);
                count_dangling_energy_right (sequence, link, i, f[i].pair, f[i].bri[f[i].num_branches-1], f[f[i].bri[f[i].num_branches-1]].pair, counter);
            }
            // add "no-dangling" restriction                                    
            for (l=0; l < f[i].num_branches; l++)
            {
                cannot_add_dangling [f[i].bri[l] -1] = 1;
                cannot_add_dangling [f[f[i].bri[l]].pair + 1] = 1;
            }
            if (debug)
            {
                printf ("%d - multi m\t- add energy %6d\n", i, misc_energy);            
                printf ("%d - multi d\t- add energy %6d\n", i, dang);
                printf ("%d - multi AU\t- add energy %6d\n", i, AUpen);
            }                
            energy += misc_energy + dang + AUpen;                           
        }
    }
    //    return energy;
}



void count_each_structure_type (char *sequence, char *structure, double *counter, int reset)
// Mirela: Nov 23, 2003
// PRE: string_params have been filled, i.e. by calling create_string_params() or create_building_block_strings()
// GIven the sequence and the structure, returns the vector counter, with the #of elementary structures
// The sequence and structure may have a space inside, meaning duplex
// If reset is 1, then counter will be reset to 0 first, otherwise, the values will be added to the current value of counter
{
    int i, j;
    str_features *f;
    int *int_sequence;
    char *actual_seq;
    char *actual_str;
    int len, nb_nucleotides;
    char *space;
    int link;
    
    /*
    for (i=0; i < NUM_DANG; i++)
    {
        for (j=0; j < NUM_DANG; j++)
        {
            counter_min_dangle[i][j] = 0;
        }
    }
    */    

    if (reset)
    {
        //first reset counter
        // counter[num_params] is the free value
        for (i=0; i <= num_params; i++)
        {
            counter[i] = 0;
        }
    }
            
    len = strlen(sequence);
    // look for the space
    space = strstr (sequence, " ");
    if (space == NULL)    // single sequence
    {
        // make sure the structure doesn't have space either
        if (strstr (structure, " ") != NULL) 
        { 
            printf ("Structure has space, and sequence not!!\n%s\n%s\n", sequence, structure); 
            exit (1);
        }
        actual_seq = sequence;
        actual_str = structure;
        link = -1;            
    }
    else
    {
        // make sure structure has space at the same place
        int space_location;
        space_location = space-sequence;
        link = space_location - 1;
        if (structure[space_location] != ' ')
        {
            printf ("Structure doesn't have space at the same place as sequence!!\n%s\n%s\n", sequence, structure); 
            exit (1);
        }
        actual_seq = new char[len+1];
        actual_str = new char[len+1];
        for (i=0; i < space_location; i++)
        {
            actual_seq[i] = sequence[i];
            actual_str[i] = structure[i];
        }
        for (i=space_location+1; i < len; i++)
        {
            actual_seq[i-1] = sequence[i];
            actual_str[i-1] = structure[i];
        }
        actual_seq[len-1] = '\0';
        actual_str[len-1] = '\0';
    }
    
    nb_nucleotides = strlen(actual_seq);    
    if ((f = new str_features[nb_nucleotides]) == NULL) giveup ("Cannot allocate memory", "str_features");    
    // detect the structure features
    // this function call should be the same for single sequence of duplex
    detect_structure_features (actual_str, f);
    
    int_sequence = new int[nb_nucleotides];
    if (int_sequence == NULL) giveup ("Cannot allocate memory", "energy");
    for (i=0; i < nb_nucleotides; i++) int_sequence[i] = nuc_to_int(actual_seq[i]);
    count_types (link, int_sequence, actual_seq, f, counter);
    if (link > -1)
    {
        delete [] actual_seq;
        delete [] actual_str;
    }    
    delete [] int_sequence;
    delete [] f;    
}



int create_building_block_strings ()
// Mirela: Sep 20, 2005
// For most of the building blocks, write which sequence(s) correspond to each building block
// also fill string_params
{
  int index;
  int i, j, k, l, m, n, o, p;
  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                // TODO: to come back to duplicates
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    sprintf (string_params[index], "stack[%d][%d][%d][%d]", i, j, k, l);
                    bbseq_left [index][0] = int_to_nuc (i);
                    bbseq_left [index][1] = int_to_nuc (k);
                    bbseq_left [index][2] = '\0';
                    strcpy (bbstr_left[index], "(("); bbstr_left[index][2] = '\0';
                    bbseq_right[index][0] = int_to_nuc (l);
                    bbseq_right[index][1] = int_to_nuc (j);
                    bbseq_right[index][2] = '\0';
                    strcpy (bbstr_right[index], "))"); bbstr_right[index][2] = '\0';
                    index++;
                  }
              }
          }          
          
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index], "tstackh[%d][%d][%d][%d]", i, j, k, l);
                bbseq_left [index][0] = int_to_nuc (i);
                bbseq_left [index][1] = int_to_nuc (k);
                bbseq_left [index][2] = '\0';
                strcpy (bbstr_left[index], "(."); bbstr_left[index][2] = '\0';
                bbseq_right[index][0] = int_to_nuc (l);
                bbseq_right[index][1] = int_to_nuc (j);
                bbseq_right[index][2] = '\0';
                strcpy (bbstr_right[index], ".)"); bbstr_right[index][2] = '\0';
                index++;
              }
          }

    // replace all tstacki parameters by 3 parameters, as described in Mathews 1999
    sprintf (string_params[index], "misc.internal_AU_closure");
    bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
    sprintf (string_params[index], "misc.internal_AG_mismatch");
    bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
    sprintf (string_params[index], "misc.internal_UU_mismatch");
    bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
    

/*          
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index], "tstacki[%d][%d][%d][%d]", i, j, k, l);
                // here we have to be careful: this is only half of an internal loop
                bbseq_left [index][0] = int_to_nuc (i);
                bbseq_left [index][1] = int_to_nuc (k);
                bbseq_left [index][2] = '\0';
                strcpy (bbstr_left[index], "(."); bbstr_left[index][2] = '\0';
                bbseq_right[index][0] = int_to_nuc (l);
                bbseq_right[index][1] = int_to_nuc (j);
                bbseq_right[index][2] = '\0';
                strcpy (bbstr_right[index], ".)"); bbstr_right[index][2] = '\0';
                index++;                
              }
          }
*/          
    if (!simple_internal_energy)          
    {

        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            sprintf (string_params[index], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                                            bbseq_left [index][0] = int_to_nuc (i);
                                            bbseq_left [index][1] = int_to_nuc (k);
                                            bbseq_left [index][2] = int_to_nuc (m);
                                            bbseq_left [index][3] = '\0';
                                            strcpy (bbstr_left[index], "(.("); bbstr_left[index][3] = '\0';
                                            bbseq_right[index][0] = int_to_nuc (n);
                                            bbseq_right[index][1] = int_to_nuc (l);
                                            bbseq_right[index][2] = int_to_nuc (j);
                                            bbseq_right[index][3] = '\0';
                                            strcpy (bbstr_right[index], ").)"); bbstr_right[index][3] = '\0';
                                            index++;
                                        }
                                    }
                                }
                            }
        sprintf (string_params[index], "misc.internal11_basic_mismatch");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal11_GG_mismatch");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;                                      
                            

        /*
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                sprintf (string_params[index], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                                bbseq_left [index][0] = int_to_nuc (i);
                                bbseq_left [index][1] = int_to_nuc (k);
                                bbseq_left [index][2] = int_to_nuc (m);
                                bbseq_left [index][3] = '\0';
                                strcpy (bbstr_left[index], "(.("); bbstr_left[index][3] = '\0';
                                bbseq_right[index][0] = int_to_nuc (n);
                                bbseq_right[index][1] = int_to_nuc (l);
                                bbseq_right[index][2] = int_to_nuc (j);
                                bbseq_right[index][3] = '\0';
                                strcpy (bbstr_right[index], ").)"); bbstr_right[index][3] = '\0';
                                index++;
                            }
                        }
                    }
        */    
        
        // go with few parameters, as in Mathews et al 1999
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                {
                    if (!can_pair(k,l) && !can_pair(k,o))
                    {
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            sprintf (string_params[index], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            bbseq_left [index][0] = int_to_nuc (i);
                            bbseq_left [index][1] = int_to_nuc (k);
                            bbseq_left [index][2] = int_to_nuc (m);
                            bbseq_left [index][3] = '\0';
                            strcpy (bbstr_left[index], "(.("); bbstr_left[index][3] = '\0';
                            bbseq_right[index][0] = int_to_nuc (n);
                            bbseq_right[index][1] = int_to_nuc (o);
                            bbseq_right[index][2] = int_to_nuc (l);
                            bbseq_right[index][3] = int_to_nuc (j);
                            bbseq_right[index][4] = '\0';
                            strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                            index++;
                        }
                    }
                }
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                {
                    if (!can_pair(k,l) && !can_pair(k,o))
                    {
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            sprintf (string_params[index], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            bbseq_left [index][0] = int_to_nuc (i);
                            bbseq_left [index][1] = int_to_nuc (k);
                            bbseq_left [index][2] = int_to_nuc (m);
                            bbseq_left [index][3] = '\0';
                            strcpy (bbstr_left[index], "(.("); bbstr_left[index][3] = '\0';
                            bbseq_right[index][0] = int_to_nuc (n);
                            bbseq_right[index][1] = int_to_nuc (o);
                            bbseq_right[index][2] = int_to_nuc (l);
                            bbseq_right[index][3] = int_to_nuc (j);
                            bbseq_right[index][4] = '\0';
                            strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                            index++;
                        }
                    }
                }
        sprintf (string_params[index], "misc.internal21_match");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal21_AU_closure");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
                                    
                            
        /*
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            // no duplicates here
                            sprintf (string_params[index], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            bbseq_left [index][0] = int_to_nuc (i);
                            bbseq_left [index][1] = int_to_nuc (k);
                            bbseq_left [index][2] = int_to_nuc (m);
                            bbseq_left [index][3] = '\0';
                            strcpy (bbstr_left[index], "(.("); bbstr_left[index][3] = '\0';
                            bbseq_right[index][0] = int_to_nuc (n);
                            bbseq_right[index][1] = int_to_nuc (o);
                            bbseq_right[index][2] = int_to_nuc (l);
                            bbseq_right[index][3] = int_to_nuc (j);
                            bbseq_right[index][4] = '\0';
                            strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                            index++;
                            }
                        }
        */                        

        // go with 53 parameters, like in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                sprintf (string_params[index], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                bbseq_left [index][0] = int_to_nuc (i);
                                bbseq_left [index][1] = int_to_nuc (k);
                                bbseq_left [index][2] = int_to_nuc (o);
                                bbseq_left [index][3] = int_to_nuc (m);
                                bbseq_left [index][4] = '\0';
                                strcpy (bbstr_left[index], "(..("); bbstr_left[index][4] = '\0';
                                bbseq_right[index][0] = int_to_nuc (n);
                                bbseq_right[index][1] = int_to_nuc (p);
                                bbseq_right[index][2] = int_to_nuc (l);
                                bbseq_right[index][3] = int_to_nuc (j);
                                bbseq_right[index][4] = '\0';
                                strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                                index++;
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                sprintf (string_params[index], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                bbseq_left [index][0] = int_to_nuc (i);
                                bbseq_left [index][1] = int_to_nuc (k);
                                bbseq_left [index][2] = int_to_nuc (o);
                                bbseq_left [index][3] = int_to_nuc (m);
                                bbseq_left [index][4] = '\0';
                                strcpy (bbstr_left[index], "(..("); bbstr_left[index][4] = '\0';
                                bbseq_right[index][0] = int_to_nuc (n);
                                bbseq_right[index][1] = int_to_nuc (p);
                                bbseq_right[index][2] = int_to_nuc (l);
                                bbseq_right[index][3] = int_to_nuc (j);
                                bbseq_right[index][4] = '\0';
                                strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                                index++;
                            }                        
                        }
                    }
        */
        // then add the 4 deltas
        sprintf (string_params[index], "misc.internal22_delta_same_size");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal22_delta_different_size");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal22_delta_1stable_1unstable");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal22_delta_AC");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          
        sprintf (string_params[index], "misc.internal22_match");
        bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;          

        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {
                                    sprintf (string_params[index], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                    bbseq_left [index][0] = int_to_nuc (i);
                                    bbseq_left [index][1] = int_to_nuc (k);
                                    bbseq_left [index][2] = int_to_nuc (o);
                                    bbseq_left [index][3] = int_to_nuc (m);
                                    bbseq_left [index][4] = '\0';
                                    strcpy (bbstr_left[index], "(..("); bbstr_left[index][4] = '\0';
                                    bbseq_right[index][0] = int_to_nuc (n);
                                    bbseq_right[index][1] = int_to_nuc (p);
                                    bbseq_right[index][2] = int_to_nuc (l);
                                    bbseq_right[index][3] = int_to_nuc (j);
                                    bbseq_right[index][4] = '\0';
                                    strcpy (bbstr_right[index], ")..)"); bbstr_right[index][4] = '\0';
                                    index++;
                                }
                            }
                        }
        */                        
    }    // end if (!simple_internal_energy)                        
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index], "dangle_top[%d][%d][%d]", i, j, k);
              bbseq_left [index][0] = int_to_nuc (j);
              bbseq_left [index][1] = '\0';
              strcpy (bbstr_left[index], "("); bbstr_left[index][1] = '\0';
              bbseq_right[index][0] = int_to_nuc (i);
              bbseq_right[index][1] = int_to_nuc (k);
              bbseq_right[index][2] = '\0';
              strcpy (bbstr_right[index], ")."); bbstr_right[index][2] = '\0';
              index++;
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index], "dangle_bot[%d][%d][%d]", i, j, k);
              bbseq_left [index][0] = int_to_nuc (k);
              bbseq_left [index][1] = int_to_nuc (j);
              bbseq_left [index][2] = '\0';
              strcpy (bbstr_left[index], ".("); bbstr_left[index][2] = '\0';
              bbseq_right[index][0] = int_to_nuc (i);
              bbseq_right[index][1] = '\0';
              strcpy (bbstr_right[index], ")"); bbstr_right[index][1] = '\0';
              index++;
            }
        }
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)
    {
        
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index], "internal_penalty_by_size[%d]", i);
          bbseq_left [index][0] = '\0';
          bbseq_right[index][0] = '\0';
          bbstr_left [index][0] = '\0';
          bbstr_right[index][0] = '\0';
          index++;
        }
    }
  for (i=1; i <= MAXLOOP_B; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index], "bulge_penalty_by_size[%d]", i);
          bbseq_left [index][0] = '\0';
          bbseq_right[index][0] = '\0';
          bbstr_left [index][0] = '\0';
          bbstr_right[index][0] = '\0';
          index++;          
        }
    }
  for (i=1; i <= MAXLOOP_H; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index], "hairpin_penalty_by_size[%d]", i);
          bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
        }
    }

 //sprintf (string_params[index], "misc.param_greater30");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.terminal_AU_penalty");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.hairpin_GGG");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.hairpin_c1");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.hairpin_c2");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.hairpin_c3");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 // TODO: TO ADD the following
 //sprintf (string_params[index], "misc.asymmetry_penalty_max_correction");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
 //sprintf (string_params[index], "misc.asymmetry_penalty_array[0]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
 //sprintf (string_params[index], "misc.asymmetry_penalty_array[1]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
 
 // the next ones are never used
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
 
 // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.

 //sprintf (string_params[index], "misc.asymmetry_penalty[1]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.asymmetry_penalty[2]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.asymmetry_penalty[3]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.asymmetry_penalty[4]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.asymmetry_penalty[5]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.asymmetry_penalty[6]");
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 //sprintf (string_params[index], "misc.gail_rule");    // this is fixed
 //bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.multi_offset");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.multi_helix_penalty");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.multi_free_base_penalty");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 sprintf (string_params[index], "misc.intermolecular_initiation");
 bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;

 for(i=0; i < nb_triloops; i++)
   {
     // TODO: I can do this
     sprintf (string_params[index], "triloop[%d].energy", i);
     bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
   }

 for(i=0; i < nb_tloops; i++)
   {
     sprintf (string_params[index], "tloop[%d].energy", i);
     bbseq_left [index][0] = '\0'; bbseq_right[index][0] = '\0'; bbstr_left [index][0] = '\0'; bbstr_right[index][0] = '\0'; index++;
   }
 return index;
}



int create_string_params ()
  // Mirela: Nov 23, 2003
  // writes each parameter type, excluding duplicates, in a long vector, containing the names of the parameters
{
  int index;
  int i, j, k, l, m, n, o, p;
  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    sprintf (string_params[index++], "stack[%d][%d][%d][%d]", i, j, k, l);
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }
          }
    // 3 params instead of the whole tstacki table
    sprintf (string_params[index++], "misc.internal_AU_closure");
    sprintf (string_params[index++], "misc.internal_AG_mismatch");
    sprintf (string_params[index++], "misc.internal_UU_mismatch");
    
    /*          
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
    */          
    if (!simple_internal_energy)          
    {

        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                                        }
                                    }
                                }
                            }
        sprintf (string_params[index++], "misc.internal11_basic_mismatch");
        sprintf (string_params[index++], "misc.internal11_GG_mismatch");

        /*    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                            }
                        }
                    }
        */
                    
        // go with few parameters, as in Mathews et al 1999
        // closed by CG
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                        }
        // closed by GC                        
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                        }
        sprintf (string_params[index++], "misc.internal21_match");
        sprintf (string_params[index++], "misc.internal21_AU_closure");
                    
        /*                    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            // no duplicates here
                            sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            }
                        }
        */                
        // 53 params instead of all, as in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        */
        // then add the 4 deltas
        sprintf (string_params[index++], "misc.internal22_delta_same_size");
        sprintf (string_params[index++], "misc.internal22_delta_different_size");
        sprintf (string_params[index++], "misc.internal22_delta_1stable_1unstable");
        sprintf (string_params[index++], "misc.internal22_delta_AC");
        sprintf (string_params[index++], "misc.internal22_match");

        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {
                                    sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                }
                            }
                        }
        */
    }    // end if (!simple_internal_energy)                        
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)        
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_B; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_H; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }

// sprintf (string_params[index++], "misc.param_greater30");
 sprintf (string_params[index++], "misc.terminal_AU_penalty");
 sprintf (string_params[index++], "misc.hairpin_GGG");
 sprintf (string_params[index++], "misc.hairpin_c1");
 sprintf (string_params[index++], "misc.hairpin_c2");
 sprintf (string_params[index++], "misc.hairpin_c3");
 
 // TODO
 //sprintf (string_params[index++], "misc.asymmetry_penalty_max_correction");
 //sprintf (string_params[index++], "misc.asymmetry_penalty_array[0]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty_array[1]");
 
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
 // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.
 //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
 //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");

 //sprintf (string_params[index++], "misc.gail_rule");
 sprintf (string_params[index++], "misc.multi_offset");
 sprintf (string_params[index++], "misc.multi_helix_penalty");
 sprintf (string_params[index++], "misc.multi_free_base_penalty");
 sprintf (string_params[index++], "misc.intermolecular_initiation");

 for(i=0; i < nb_triloops; i++)
   {
     sprintf (string_params[index++], "triloop[%d].energy", i);
   }

 for(i=0; i < nb_tloops; i++)
   {
     sprintf (string_params[index++], "tloop[%d].energy", i);
   }
 return index;
}


int check_stability_and_size (int k, int l, int o, int p)
// helper function, to detect which delta we need for the int22 parameters
{
    // having at least one AC mismatch is the simplest, test first
    if ((k==A && l==C) || (k==C && l==A) || (o==A && p==C) || (o==C && p==A))
        return 4;
    
    // combination of all mismatches of equal size (purine-purine, purine-pyrimidine, and pyrimidine-pyrimidine are different sizes)
    // purine =  A, G
    // pyrimidine = C, U
    // if all purine-purines
    if ((k==A || k==G) && (l==A || l==G) && (o==A || o==G) && (p==A || p==G))
        return 1;
    // if all pyrimidine-pyrimidine
    if ((k==C || k==U) && (l==C || l==U) && (o==C || o==U) && (p==C || p==U))
        return 1;
    // if both  purine-pyrimidine
    // assume the A-C pairs have been found above
    if ( (((k==A || k==G) && (l==C || l==U)) || ((k==C || k==U) && (l==A || l==G))) &&
         (((o==A || o==G) && (p==C || p==U)) || ((o==C || o==U) && (p==A || p==G))) )
        return 1;
    // or any combination of 2 unstable mismatches except AC: AA, CC, CU, GG
    if ( ((k==A && l==A) || (k==C && l==C) || (k==C && l==U) || (k==U && l==C) || (k==G && l==G)) &&
         ((o==A && p==A) || (o==C && p==C) || (o==C && p==U) || (o==U && p==C) || (o==G && p==G)) )
        return 1;
                 
    // two stabilizing mismatches (GU, GA, UU) of different sizes  (purine-purine, purine-pyrimidine, and pyrimidine-pyrimidine are different sizes)
    if ( (((k==G && l==U) || (k==U && l==G))   &&   ((o==G && p==A) || (o==A && p==G) || (o==U && p==U))) ||
         (((k==G && l==A) || (k==A && l==G))   &&   ((o==G && p==U) || (o==U && p==G) || (o==U && p==U))) ||
         ((k==U && l==U)                       &&   ((o==G && p==A) || (o==A && p==G) || (o==G && p==U) || (o==U && p==G))) )
        return 2;
        
    // one stable (GU, GA, UU) and one unstable mismatch (excluding AC) (AA, CC, CU, GG) of different sizes
    // GU
    if ( ((k==G && l==U) || (k==U && l==G)) && 
              ((o==A && p==A) || (o==C && p==C) || (o==C && p==U) || (o==U && p==C) || (o==G && p==G)) )
        return 3;    
    if ( ((o==G && p==U) || (o==U && p==G)) &&
              ((k==A && l==A) || (k==C && l==C) || (k==C && l==U) || (k==U && l==C) || (k==G && l==G)) )
        return 3;
    // GA        
    if ( ((k==G && l==A) || (k==A && l==G)) &&
              ((o==C && p==C) || (o==C && p==U) || (o==U && p==C)) )
        return 3;    
    if ( ((o==G && p==A) || (o==A && p==G)) &&
              ((k==C && l==C) || (k==C && l==U) || (k==U && l==C)) )
        return 3;
    // UU        
    if ( (k==U && l==U) &&
              ((o==A && p==A) || (o==G && p==G)) )
        return 3;    
    if ( (o==U && p==U) &&
              ((k==A && l==A) || (k==G && l==G)) )
        return 3;    
        
    return -1;            
        
}


void fill_data_structures_with_new_parameters (char *filename)
  // Mirela: Dec 16, 2003
  // reads parameters from a file, and writes them in the internal data structures
  // PRE: first read the actual standard parameters, to be able to figure out which of them are
  // < INF, and maybe to also keep some old values.
{
    int index;
    int i, j, k, l, m, n, o, p;
    FILE *file;
    char buffer[100];
    double param;
    int line = 0;

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                        {
                            int11[i][j][k][l][m][n] = INF;
                        }

    
    //printf ("FILENAME: %s\n", filename);
    if ((file = fopen (filename, "r")) == NULL)
    {
        giveup ("Cannot open file", filename);
    }
    
    index = 0;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (stack[i][j][k][l] < INF)
                {
                    // exclude duplicates
                    // stack[i][j][k][l] is the same as stack[l][k][j][i]
                    if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                    {
                        fgets (buffer, sizeof(buffer), file);
                        line++;
                        sscanf (buffer, "%lf\n", &param);
                        //printf ("\t%lf\n", param);
                        // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                        {
                            param *= 100;
                            stack[i][j][k][l] = (int)(round(param));
                            // add the duplicate too
                            stack[l][k][j][i] = (int)(round(param));
                        }
                        //sprintf (string_params[index++], "stack[%d][%d][%d][%d]", i, j, k, l);
                    }
                }
            }
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (tstackh[i][j][k][l] < INF)
                {
                    // no duplicates here
                    fgets (buffer, sizeof(buffer), file);
                    line++;
                    sscanf (buffer, "%lf", &param);
                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                        param *= 100;
                        tstackh[i][j][k][l] = (int)(round(param));
                    }
                    //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
                }
            }
    // use only 3 parameters, and fill the tstacki data structure by applying the Mathews 99 rules
    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_AU_closure = (int)(round(param));
    
    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_AG_mismatch = (int)(round(param));

    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_UU_mismatch = (int)(round(param));
    
    // fill the tstacki data structure a bit later, after we read AU_penalty
    
    /*
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (tstacki[i][j][k][l] < INF)
                {
                    // no duplicates here
                    fgets (buffer, sizeof(buffer), file);
                    sscanf (buffer, "%lf", &param);
                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                        param *= 100;
                        tstacki[i][j][k][l] = (int)(round(param));
                    }
                    //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
                }
            }
    */            
    if (!simple_internal_energy)
    {
        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    //if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            fgets (buffer, sizeof(buffer), file);
                                            line++;
                                            sscanf (buffer, "%lf", &param);
                                            param *= 100;
                                            int11[i][j][k][l][m][n] = (int)(round(param));
                                            // add the duplicate too
                                            int11[n][m][l][k][j][i] = (int)(round(param));
                                        }
                                    }
                                }
                            }
        fgets (buffer, sizeof(buffer), file);
        line++;
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal11_basic_mismatch = (int)(round(param));
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal11_GG_mismatch = (int)(round(param));


        // fill the int11 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        if ( ((i==C && j==G) || (i==G && j==C)) &&
                                             ((m==C && n==G) || (m==G && n==C)))
                                        {
                                            if (can_pair(k,l))
                                            {
                                                int11[i][j][k][l][m][n] = misc.internal11_basic_mismatch;
                                                // add the duplicate too
                                                //int11[n][m][l][k][j][i] = misc.internal11_basic_mismatch;
                                            }
                                        }
                                        
                                        else
                                        {
                                            if (!(watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                            {
                                                if (k==G && l==G)
                                                {
                                                    int11[i][j][k][l][m][n] = misc.internal11_GG_mismatch;
                                                    // add the duplicate too
                                                    //int11[n][m][l][k][j][i] = misc.internal11_GG_mismatch;
                                                }
                                                else
                                                    int11[i][j][k][l][m][n] = misc.internal11_basic_mismatch;
                                                if (has_AU_penalty(i,j))
                                                    int11[i][j][k][l][m][n] += misc.internal_AU_closure;
                                                if (has_AU_penalty(m,n))
                                                    int11[i][j][k][l][m][n] += misc.internal_AU_closure;
                                            }
                                        }
                                        
                                        // round it to match Turner parameters
                                        //if (int11[i][j][k][l][m][n] % 10 == 5) int11[i][j][k][l][m][n] += 5;
                                        //if (int11[i][j][k][l][m][n] % 10 == -5) int11[i][j][k][l][m][n] += 5;
                                    }
                }                                

        /*    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int11[i][j][k][l][m][n] = (int)(round(param));
                                    // add the duplicate too
                                    int11[n][m][l][k][j][i] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                            }
                        }
                    }
        */
                            
        // go with few int21 parameters, as in Mathews et al 1999
        // closed by CG
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            param *= 100;
                            int21[i][j][k][l][m][n][o] = (int)(round(param));
                        }
        // closed by GC                        
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            param *= 100;
                            int21[i][j][k][l][m][n][o] = (int)(round(param));
                        }
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal21_match = (int)(round(param));
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal21_AU_closure = (int)(round(param));
       
        // fill the int21 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        for(o=0; o < NUCL; o++)
                                        {
                                            if ((i==C && j==G && m==C && n==G) ||  // these are already filled above, except what can pair inside
                                                (i==G && j==C && m==G && n==C))
                                            {
                                                if (can_pair(k,l) || can_pair(k,o))
                                                    int21[i][j][k][l][m][n][o] = misc.internal21_match;
                                            }
                                            else
                                            {
                                                if (can_pair(k,l) || can_pair(k,o))
                                                    int21[i][j][k][l][m][n][o] = misc.internal21_match;
                                                // I approximate it to int, so that we obtain the same as by counte_each_structure_type and summing up
                                                else
                                                    int21[i][j][k][l][m][n][o] = (int)(int21[C][G][k][l][C][G][o]/2) +
                                                                                 (int)(int21[G][C][k][l][G][C][o]/2);
                                                if (has_AU_penalty(i,j))
                                                    int21[i][j][k][l][m][n][o] += misc.internal21_AU_closure;
                                                if (has_AU_penalty(m,n))    
                                                    int21[i][j][k][l][m][n][o] += misc.internal21_AU_closure;
                                            }
                                            // round it to match Turner parameters - seems to be inconsistent
                                            //if (int21[i][j][k][l][m][n][o] % 10 == 5) int21[i][j][k][l][m][n][o] += 5;
                                        }
                                    }
                }                                
        

                    
        /*                    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                param *= 100;
                                int21[i][j][k][l][m][n][o] = (int)(round(param));
                                }
                            //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            }
                        }
        */                        
        // go with the 53 parameters, like in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                    // add the duplicate too
                                    int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                    // add the duplicate too
                                    int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        */            
        // then add the 4 deltas
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_same_size = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_different_size = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_1stable_1unstable = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_AC = (int)(round(param));

        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_match = (int)(round(param));

        int ii, jj, mm, nn;
        // fill the int22 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        for(o=0; o < NUCL; o++)
                                            for (p=0; p < NUCL; p++)
                                            {
                                                /*
                                                if (i==C && j==G && m==C && n==G)
                                                {
                                                    if(watson_crick(k,l) || watson_crick(o,p))
                                                    {
                                                        int22[i][j][k][l][m][n][o][p] = misc.internal22_match;
                                                    }
                                                    // else do nothing, it's parameter
                                                } */
                                                // if a closing pair is wobble, it's the same as if G would be A
                                                if (i==G && j==U)   ii = A;     else ii = i;
                                                if (i==U && j==G)   jj = A;     else jj = j;
                                                if (m==G && n==U)   mm = A;     else mm = m;
                                                if (m==U && n==G)   nn = A;     else nn = n;

                                                
                                                if (watson_crick(k,l) || watson_crick(o,p))
                                                {
                                                    int22[i][j][k][l][m][n][o][p] = misc.internal22_match;
                                                }
                                                else if (nn==ii && mm==jj && p==k && o==l)  // the UG closing pairs are the same as UA
                                                {
                                                    int22[i][j][k][l][m][n][o][p] = int22[ii][jj][k][l][mm][nn][o][p];
                                                }
                                                else //if (!(n==i && m==j && p==k && o==l))   // was already filled above
                                                {
                                                    int result = check_stability_and_size (k, l, o, p);
                                                    // I approximate it to int, so that we obtain the same as by counte_each_structure_type and summing up
                                                    int temp = (int)(int22[ii][jj][k][l][jj][ii][l][k]/2.0) +
                                                               (int)(int22[nn][mm][p][o][mm][nn][o][p]/2.0);
                                                    // rounf it to match Turner parameters
                                                    //if (temp%10 == 5) temp -= 5; if (temp%10 == -5) temp += 5;
                                                    int22[i][j][k][l][m][n][o][p] =  temp;
                                                    switch (result)
                                                    {
                                                        case 1: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_same_size; break;
                                                        case 2: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_different_size; break;
                                                        case 3: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_1stable_1unstable; break;
                                                        case 4: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_AC; break;
                                                        default: printf ("ERROR: result %d for k=%d, l=%d, o=%d, p=%d, ABORT!\n", result, k, l, o, p); exit(1);                                                
                                                    }                                                
                                                }                                        
                                            }
                                    }
                }                                
                        
        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {
                                    fgets (buffer, sizeof(buffer), file);
                                    sscanf (buffer, "%lf", &param);
                                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                    {
                                        param *= 100;
                                        int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                        // add the duplicate too
                                        int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                    }
                                    //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                }
                            }
                        }
        */                    
    }     // end if (!simple_internal_energy)
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
            {
            if (dangle_top[i][j][k] < INF)
                {
                // no duplicates here
                fgets (buffer, sizeof(buffer), file);
                sscanf (buffer, "%lf", &param);
                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                    param *= 100;
                    dangle_top[i][j][k] = (int)(round(param));
                    }
                //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
                }
            }
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            {
            if (dangle_bot[i][j][k] < INF)
                {
                // no duplicates here
                fgets (buffer, sizeof(buffer), file);
                sscanf (buffer, "%lf", &param);
                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                    param *= 100;
                    dangle_bot[i][j][k] = (int)(round(param));
                    }
                //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
                }
            }
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)
        {
        if (internal_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                internal_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
            }
        }
    for (i=1; i <= MAXLOOP_B; i++)
        {
        if (bulge_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);    
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                bulge_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
            }
        }
    for (i=1; i <= MAXLOOP_H; i++)
        {
        if (hairpin_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                hairpin_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
            }
        }
    //fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
    //    misc.param_greater30 = (int)(round(param));
    
    // set a fixed value to param_greater_30 for now
    misc.param_greater30 = 1.079;
    
    //sprintf (string_params[index++], "misc.param_greater30");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.terminal_AU_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.terminal_AU_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_GGG = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_GGG");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c1 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c1");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c2 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c2");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c3 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c3");

    // fill the tstacki data structure, now that we also have terminal_AU_penalty - actually I removed it from here
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (!can_pair (i, j))
                        tstacki[i][j][k][l] = INF;
                    else    
                    {
                        tstacki[i][j][k][l] = 0;
                        if (((i == A || i == G) && j == U) ||
                            ((j == A || j == G) && i == U))
                        {
                            tstacki[i][j][k][l] += misc.internal_AU_closure;
                        }
                        if ((k == A && l == G) ||
                            (l == A && k == G))
                        {
                            tstacki[i][j][k][l] += misc.internal_AG_mismatch;
                        }
                        if (k == U && l == U)
                        {
                            tstacki[i][j][k][l] += misc.internal_UU_mismatch;
                        }
                    }
                }

    
    // TODO
    // keep them fixed for now    
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); 
    // sprintf (string_params[index++], "misc.asymmetry_penalty_max_correction");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[0]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[1]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
    // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.
    
    // to come back!!!
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");
    
    //fgets (buffer, sizeof(buffer), file);
    //sprintf (string_params[index++], "misc.gail_rule");
    // keep this fixed 
    misc.gail_rule = 1;
    
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_offset = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_offset");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_helix_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_helix_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_free_base_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_free_base_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.intermolecular_initiation = (int)(round(param));
    //sprintf (string_params[index++], "misc.intermolecular_initiation");
    
    for(i=0; i < nb_triloops; i++)
        {
        fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
        // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
            triloop[i].energy = (int)(round(param));
        //sprintf (string_params[index++], "triloop[%d].energy", i);
        }
    
    for(i=0; i < nb_tloops; i++)
        {
        fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
        // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
            tloop[i].energy = (int)(round(param));
        //sprintf (string_params[index++], "tloop[%d].energy", i);
        }
    
    fclose (file);
    //printf ("****** stack[1][2][0][3] = %d\n", stack[1][2][0][3]);
}


void save_paramtypes (char *filename)
// PRE: call create_string_params ()
// save all parameter types in the given file
{
    int index;
    int i, j, k, l, m, n, o, p;
    FILE *file;

    if ((file = fopen (filename, "w")) == NULL)
    {
        giveup ("Cannot open file", filename);
    }
    for (i=0; i < num_params; i++)
    {
        fprintf (file, "%s\n", string_params[i]);
    }
    
    fclose (file);
            
}


void save_parameters (char *filename)
  // Mirela: Dec 16, 2003
  // save all parameters in the given file
{
  int index;
  int i, j, k, l, m, n, o, p;
  FILE *file;

  if ((file = fopen (filename, "w")) == NULL)
    {
      giveup ("Cannot open file", filename);
    }

  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    fprintf (file, "%.2lf\n", stack[i][j][k][l]/100.0);
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                fprintf (file, "%.2lf\n", tstackh[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }              
          }
    // we only use 3 parameters here, instead of the whole tstacki table
    fprintf (file, "%.2lf\n", misc.internal_AU_closure/100.0);
    fprintf (file, "%.2lf\n", misc.internal_AG_mismatch/100.0);
    fprintf (file, "%.2lf\n", misc.internal_UU_mismatch/100.0);

/*          
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                fprintf (file, "%.2lf\n", tstacki[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
*/          
    if (!simple_internal_energy)          
    {
        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            fprintf (file, "%.2lf\n", int11[i][j][k][l][m][n]/100.0);
                                        }
                                    }
                                }
                            }
        fprintf (file, "%.2lf\n", misc.internal11_basic_mismatch/100.0);
        fprintf (file, "%.2lf\n", misc.internal11_GG_mismatch/100.0);
    
        /*
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                fprintf (file, "%.2lf\n", int11[i][j][k][l][m][n]/100.0);
                                //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                            }
                        }
                    }
        */                    
        // go with few parameters, as in Mathews et al 1999
        // closed by CG
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fprintf (file, "%.2lf\n", int21[i][j][k][l][m][n][o]/100.0);
                        }
        // closed by GC                        
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fprintf (file, "%.2lf\n", int21[i][j][k][l][m][n][o]/100.0);
                        }
        fprintf (file, "%.2lf\n", misc.internal21_match/100.0);
        fprintf (file, "%.2lf\n", misc.internal21_AU_closure/100.0);
        /*                    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            // no duplicates here
                            fprintf (file, "%.2lf\n", int21[i][j][k][l][m][n][o]/100.0);
                            //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            }
                        }
        */
                                
        // go with the 53 params, like in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;                    
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        */
        // add the 4 deltas
        fprintf (file, "%.2lf\n", misc.internal22_delta_same_size/100.0);
        fprintf (file, "%.2lf\n", misc.internal22_delta_different_size/100.0);
        fprintf (file, "%.2lf\n", misc.internal22_delta_1stable_1unstable/100.0);
        fprintf (file, "%.2lf\n", misc.internal22_delta_AC/100.0);
        fprintf (file, "%.2lf\n", misc.internal22_match/100.0);
        
        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {                          
                                    fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                    //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                }
                            }
                        }
        */                        
    }    // end if (!simple_internal_energy)                        
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              fprintf (file, "%.2lf\n", dangle_top[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              fprintf (file, "%.2lf\n", dangle_bot[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", internal_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_B; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", bulge_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_H; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", hairpin_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }
  //fprintf (file, "%.2lf\n", misc.param_greater30/100.0);
  //sprintf (string_params[index++], "misc.param_greater30");
  fprintf (file, "%.2lf\n", misc.terminal_AU_penalty/100.0);
  //sprintf (string_params[index++], "misc.terminal_AU_penalty");
  fprintf (file, "%.2lf\n", misc.hairpin_GGG/100.0);
  //sprintf (string_params[index++], "misc.hairpin_GGG");
  fprintf (file, "%.2lf\n", misc.hairpin_c1/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c1");
  fprintf (file, "%.2lf\n", misc.hairpin_c2/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c2");
  fprintf (file, "%.2lf\n", misc.hairpin_c3/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c3");
  
  // TODO
  //fprintf (file, "%d\n", misc.asymmetry_penalty_max_correction);
  //fprintf (file, "%d\n", misc.asymmetry_penalty_array[0]);
  //fprintf (file, "%d\n", misc.asymmetry_penalty_array[1]);
  
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");  
  
  
  //fprintf (file, "%.2lf\n", misc.gail_rule);
  //sprintf (string_params[index++], "misc.gail_rule");
  fprintf (file, "%.2lf\n", misc.multi_offset/100.0);
  //sprintf (string_params[index++], "misc.multi_offset");
  fprintf (file, "%.2lf\n", misc.multi_helix_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_helix_penalty");
  fprintf (file, "%.2lf\n", misc.multi_free_base_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_free_base_penalty");
  fprintf (file, "%.2lf\n", misc.intermolecular_initiation/100.0);
  //sprintf (string_params[index++], "misc.intermolecular_initiation");

 for(i=0; i < nb_triloops; i++)
   {
     fprintf (file, "%.2lf\n", triloop[i].energy/100.0);
     //sprintf (string_params[index++], "triloop[%d].energy", i);
   }

 for(i=0; i < nb_tloops; i++)
   {
     fprintf (file, "%.2lf\n", tloop[i].energy/100.0);
     //sprintf (string_params[index++], "tloop[%d].energy", i);
   }

 fclose (file);
}


void save_parameters_in_array (int *array)
// PRE: parameters have been read
// save all parameters in the given array
{
  int index;
  int i, j, k, l, m, n, o, p;

  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    array[index++] = stack[i][j][k][l];
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                array[index++] = tstackh[i][j][k][l];
                //fprintf (file, "%.2lf\n", tstackh[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }
          }
    // use only 3 parameters, instead of the tstacki table
    array[index++] = misc.internal_AU_closure;
    array[index++] = misc.internal_AG_mismatch;
    array[index++] = misc.internal_UU_mismatch;
        
        /*
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                array[index++] = tstacki[i][j][k][l];
                //fprintf (file, "%.2lf\n", tstacki[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
        */          
    if (!simple_internal_energy)          
    {

        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            array[index++] = int11[i][j][k][l][m][n];
                                        }
                                    }
                                }
                            }
        array[index++] = misc.internal11_basic_mismatch;
        array[index++] = misc.internal11_GG_mismatch;
    
        /*
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                array[index++] = int11[i][j][k][l][m][n];
                                //fprintf (file, "%.2lf\n", int11[i][j][k][l][m][n]/100.0);
                                //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                            }
                        }
                    }
        */                    
        // go with few parameters, as in Mathews et al 1999
        // closed by CG
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            array[index++] = int21[i][j][k][l][m][n][o];
                        }
        // closed by GC                        
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            array[index++] = int21[i][j][k][l][m][n][o];
                        }
        array[index++] = misc.internal21_match;
        array[index++] = misc.internal21_AU_closure;
        
        /*
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            array[index++] = int21[i][j][k][l][m][n][o];
                            // no duplicates here
                            //fprintf (file, "%.2lf\n", int21[i][j][k][l][m][n][o]/100.0);
                            //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            }
                        }
        */
                                
        // go with the 53 parameters, like in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;                                        
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                array[index++] = int22[i][j][k][l][m][n][o][p];
                                //fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                array[index++] = int22[i][j][k][l][m][n][o][p];
                                //fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        */            
        // add the 4 deltas
        array[index++] = misc.internal22_delta_same_size;
        array[index++] = misc.internal22_delta_different_size;
        array[index++] = misc.internal22_delta_1stable_1unstable;
        array[index++] = misc.internal22_delta_AC;
        array[index++] = misc.internal22_match;
                    
                        
        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {                          
                                    array[index++] = int22[i][j][k][l][m][n][o][p];
                                    //fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                                    //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                }
                            }
                        }
        */                        
    }    // end if (!simple_internal_energy)                        
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              array[index++] = dangle_top[i][j][k];
              //fprintf (file, "%.2lf\n", dangle_top[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              array[index++] = dangle_bot[i][j][k];
              //fprintf (file, "%.2lf\n", dangle_bot[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          array[index++] = internal_penalty_by_size[i];
          //fprintf (file, "%.2lf\n", internal_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_B; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          array[index++] = bulge_penalty_by_size[i];
          //fprintf (file, "%.2lf\n", bulge_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP_H; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          array[index++] = hairpin_penalty_by_size[i];
          //fprintf (file, "%.2lf\n", hairpin_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }
  //fprintf (file, "%.2lf\n", misc.param_greater30/100.0);
  //sprintf (string_params[index++], "misc.param_greater30");
  
  array[index++] = misc.terminal_AU_penalty;
  //fprintf (file, "%.2lf\n", misc.terminal_AU_penalty/100.0);
  //sprintf (string_params[index++], "misc.terminal_AU_penalty");
  
  array[index++] = misc.hairpin_GGG;
  //fprintf (file, "%.2lf\n", misc.hairpin_GGG/100.0);
  //sprintf (string_params[index++], "misc.hairpin_GGG");
  
  array[index++] = misc.hairpin_c1;
  //fprintf (file, "%.2lf\n", misc.hairpin_c1/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c1");
  
  array[index++] = misc.hairpin_c2;
  //fprintf (file, "%.2lf\n", misc.hairpin_c2/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c2");
  
  array[index++] = misc.hairpin_c3;
  //fprintf (file, "%.2lf\n", misc.hairpin_c3/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c3");
  
  // TODO
  //fprintf (file, "%d\n", misc.asymmetry_penalty_max_correction);
  //fprintf (file, "%d\n", misc.asymmetry_penalty_array[0]);
  //fprintf (file, "%d\n", misc.asymmetry_penalty_array[1]);
  
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
  //fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");  
  
  
  //fprintf (file, "%.2lf\n", misc.gail_rule);
  //sprintf (string_params[index++], "misc.gail_rule");
  
  array[index++] = misc.multi_offset;
  //fprintf (file, "%.2lf\n", misc.multi_offset/100.0);
  //sprintf (string_params[index++], "misc.multi_offset");
  
  array[index++] = misc.multi_helix_penalty;
  //fprintf (file, "%.2lf\n", misc.multi_helix_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_helix_penalty");
  
  array[index++] = misc.multi_free_base_penalty;
  //fprintf (file, "%.2lf\n", misc.multi_free_base_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_free_base_penalty");
  
  array[index++] = misc.intermolecular_initiation;
  //fprintf (file, "%.2lf\n", misc.intermolecular_initiation/100.0);
  //sprintf (string_params[index++], "misc.intermolecular_initiation");

 for(i=0; i < nb_triloops; i++)
   {
     array[index++] = triloop[i].energy;       
     //fprintf (file, "%.2lf\n", triloop[i].energy/100.0);
     //sprintf (string_params[index++], "triloop[%d].energy", i);
   }

 for(i=0; i < nb_tloops; i++)
   {
     array[index++] = tloop[i].energy;
     //fprintf (file, "%.2lf\n", tloop[i].energy/100.0);
     //sprintf (string_params[index++], "tloop[%d].energy", i);
   }

   if (index != num_params)
   {
       printf ("index=%d should be the same as num_params=%d in save_parameters_in_array\n", index, num_params);
       exit(0);
   }

}




double simfold_restricted_logZ (char *sequence, char *real_structure, char *restricted, double &min_energy, double &max_energy, int &actual_num_str)
// if the real_structure is not one of the suboptimal structures, add it 
{
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double Z;
    int i;
    double R, temp, beta;
    int real_str_found;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    real_str_found = 0;
    
    
    s_min_folding *min_fold = new s_min_folding (sequence, restricted);
    min_energy = min_fold->s_simfold_restricted ();
    min_fold->return_structure (structure);
    delete min_fold;      
    
    s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold_restricted (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    delete sub_fold;
    
    max_energy = tmp_energies[actual_num_str-1];
    Z = 0;
    for (i=0; i < actual_num_str; i++)
    {
        // check if the real structure is one of the suboptimal ones
        if (strcmp (tmp_structures[i], real_structure) == 0)
            real_str_found = 1;
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold_restricted (sequence, tmp_structures[i], restricted);
        Z += exp ((-1) * energy * beta);
    }
    if (!real_str_found)
    {
        // if not found, add it
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);
        Z += exp ((-1) * energy * beta);        
    }        
    return log(Z);    
}


double simfold_restricted_logZ_gradient (char *sequence, char *real_structure, char *restricted, double *logZ_gradient)
// return 
// if the real_structure is not one of the suboptimal structures, add it 
{
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double min_energy, max_energy;
    int i, k;
    double R, temp, beta;
    int real_str_found, actual_num_str;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    real_str_found = 0;
    double numerator [MAXNUMPARAMS];
    double counter [MAXNUMPARAMS];
    double denominator;    
    
    s_min_folding *min_fold = new s_min_folding (sequence, restricted);
    min_energy = min_fold->s_simfold_restricted ();
    min_fold->return_structure (structure);
    delete min_fold;      
    
    s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold_restricted (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    max_energy = tmp_energies[actual_num_str-1];
    delete sub_fold;
    
    for (i=0; i < num_params; i++)
    {
        numerator[i] = 0;
    }    
        
    // first compute the denominator, which is the same for all parameters
    denominator = 0;
    for (k=0; k < actual_num_str; k++)
    {
        // check if the real structure is one of the suboptimal ones
        if (strcmp (tmp_structures[k], real_structure) == 0)
            real_str_found = 1;
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold_restricted (sequence, tmp_structures[k], restricted);        
        denominator += exp ((-1) * energy * beta);
        count_each_structure_type (sequence, tmp_structures[k], counter, 1);
        for (i=0; i < num_params; i++)
        {
            numerator[i] += counter[i] * exp ((-1) * energy * beta);
        }
        //printf ("Sub str %d, counter[i] = %d, numerator[i] = %e\n", k, counter[7646], numerator[7646]);
    }
    if (!real_str_found)
    {
        //printf ("Real str not found, add it\n");
        // if not found, add it
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);
        denominator += exp ((-1) * energy * beta);        
        count_each_structure_type (sequence, real_structure, counter, 1);
        for (i=0; i < num_params; i++)
        {
            numerator[i] += counter[i] * exp ((-1) * energy * beta);
        }    
    }
    // now the denominator and nominator are computed
    
    for (i=0; i < num_params; i++)
    {
        logZ_gradient[i] = numerator[i] / denominator;
    }      
    //printf ("logZ_gradient[i] = %e\n", logZ_gradient[7646]);
}


double compute_probability (double energy, double Z)
{
    double pb;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    pb = exp ((-1) * energy * beta) / Z;
    return pb;
}

int get_info_from_file (FILE *file, char *sequence, char *real_structure, char *restricted)
// return 1 if everything was ok
// return 0 instead of continue;
{
    int MLformat = 0;
    char buffer [5000];
    int i;

    if (MLformat)
    {
        if (fgets (buffer, sizeof(buffer), file) == NULL)
            return 0;
        if (strstr (buffer, "Seq: ") != NULL)
        {
            for (i=5; i < strlen (buffer)-1; i++)
                sequence[i-5] = buffer[i]; 
            sequence[i-5] = '\0';
            // ignore sequences longer than MAXSLEN -1 
            if (strlen (sequence) > MAXSLEN-1) return 0;
        }
        else return 0;
        fgets (buffer, sizeof(buffer), file);
        if (strstr (buffer, "Str: ") != NULL)
        {
            for (i=5; i < strlen (buffer)-1; i++)
                real_structure[i-5] = buffer[i]; 
            real_structure[i-5] = '\0';
        }        
        else
        {
            printf ("Str doesn't follow Seq, Seq is %s\n", sequence);
            exit(1);
        }
        fgets (buffer, sizeof(buffer), file);
        if (strstr (buffer, "Res: ") != NULL)
        {
            for (i=5; i < strlen (buffer)-1; i++)
                restricted[i-5] = buffer[i]; 
            restricted[i-5] = '\0';
        }        
        else
        {
            printf ("Res doesn't follow Str\n");
            exit(1);
        }        
        fgets (buffer, sizeof(buffer), file);   //====
    }
    else
    {
        if (fgets (buffer, sizeof(buffer), file) == NULL)
            return 0;
        if (buffer[0] == '>')
        {
            // next should be sequence
            fgets (buffer, sizeof(buffer), file);
            for (i=0; i < strlen (buffer)-1; i++)
                sequence[i] = buffer[i];
            sequence[i] = '\0';
            
            // next should be real structure
            fgets (buffer, sizeof(buffer), file);
            for (i=0; i < strlen (buffer)-1; i++)
                real_structure[i] = buffer[i];
            real_structure[i] = '\0';

            // next is restricted or empty line
            fgets (buffer, sizeof(buffer), file);
            if (buffer[0] == '\n')
            {
                restricted[0] = '\0';
                return 1;   // that's it, we are done
            }
            for (i=0; i < strlen (buffer)-1; i++)
                restricted[i] = buffer[i];
            restricted[i] = '\0';
            fgets (buffer, sizeof(buffer), file);   // empty line
        }
        else return 0;        
    }
    return 1;
}


double compute_f (char *input_file)
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double f; 
    double min_energy, max_energy;      
    int i, actual_num_str;
        
    f = 0;
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }
      
    int k;
    k = 0;
    int seen = 0;
    while (!feof (file))
    {
        seen++;
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        //printf ("      ....,....1....,....2....,....3....,....4....,....5\n");
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);        
        logZ = simfold_restricted_logZ (sequence, real_structure, restricted, min_energy, max_energy, actual_num_str);
        //printf ("logZ(%d) = %e\n", seen, logZ);
        f += beta*energy + logZ;
    }      
    fclose (file);
    return f;                                                                                                                                
}

long double compute_likelihood_exactly (char *input_file)
// created on Oct 7, 2006
// PRE:  the parameters have been read
// POST: computes the likelihood function, which is \prod_{i=1}^N (exp (1/RT G_{i,nat,theta}) / Z)
// the partition function is computed exactly.
// the restriction string is not included so far
{

    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    FILE *file;
    long double Z;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    long double f; 
                
    f = 1.0;
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }
    
    //long double pf;
    //pf = simfold_partition_function_exactly ("CAAAAGUCUGGGCUAAGCCCACUGAUGAGCCGCUGAAAUGCGGCGAAACUUUUG");
    //printf ("Exact part fun: %Le\n", pf);    
         
    int k;
    k = 0;
    int seen = 0;
    while (!feof (file))
    {
        seen++;
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        //printf ("      ....,....1....,....2....,....3....,....4....,....5\n");
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        
        
        // for now ignore restricted if it exists
        energy = free_energy_simfold (sequence, real_structure);  
                
        Z = simfold_partition_function_smart (sequence);
        //printf ("Z = %Le\n", Z);
        //printf ("numerator = %e\n", exp (-1.0*beta*energy));
        f *= exp (-1.0*beta*energy) / Z;
    }      
    fclose (file);
    return f;                                                                                                                                
}


long double compute_log_likelihood_smart (char *input_file)
// created on Oct 7, 2006
// PRE:  the parameters have been read
// POST: computes the likelihood function, which is \prod_{i=1}^N (exp (1/RT G_{i,nat,theta}) / Z)
// the partition function is computed exactly - with teh smart McCaskill's dynamic programming algorithm.
// the restriction string is not included so far
{

    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    FILE *file;
    double Z;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double f;
                
    f = 0.0;
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }
    
    int k;
    k = 0;
    int seen = 0;
    while (!feof (file))
    {
        seen++;
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        //printf ("      ....,....1....,....2....,....3....,....4....,....5\n");
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        
        
        // for now ignore restricted if it exists
        energy = free_energy_simfold (sequence, real_structure);  
                
        Z = simfold_partition_function_smart (sequence);
        //printf ("Z = %g, en = %g, f = %g  %s\n", Z, energy, 1.0*beta*energy + log(Z), sequence);
        //printf ("numerator = %e\n", exp (-1.0*beta*energy));
        f += 1.0*beta*energy + log(Z);
    }      
    fclose (file);
    return f;                                                                                                                                
}



void compute_gradient_f (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    double counter [MAXNUMPARAMS];    
    double logZ_gradient [MAXNUMPARAMS];
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;    
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter, 1);
        simfold_restricted_logZ_gradient (sequence, real_structure, restricted, logZ_gradient);
        for (i = 0; i < num_params; i++)
        {
            f_gradient[i] += counter[i] - logZ_gradient[i];
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], logZ_gradient[7621], f_gradient[7621]);
    }      
    fclose (file);
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }    
}


void compute_gradient_f_smart (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
// unrestricted for now
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    double counter [MAXNUMPARAMS];    
    double logZ_gradient [MAXNUMPARAMS];
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter, 1);
        simfold_gradient_smart (sequence, logZ_gradient);        
        for (i = 0; i < num_params; i++)
        {
            f_gradient[i] += counter[i] - logZ_gradient[i];
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], logZ_gradient[7621], f_gradient[7621]);
    }      
    fclose (file);
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }    
}



double compute_f_and_gradient_f_smart (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
// returns f
// unrestricted for now
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    double counter [MAXNUMPARAMS];    
    double logZ_gradient [MAXNUMPARAMS];
    double Z, f;
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter, 1);
        energy = free_energy_simfold (sequence, real_structure); 
        Z = simfold_f_and_gradient_smart (sequence, logZ_gradient);
        f += 1.0*beta*energy + log(Z);
        for (i = 0; i < num_params; i++)
        {
            f_gradient[i] += counter[i] - logZ_gradient[i];
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], logZ_gradient[7621], f_gradient[7621]);
    }      
    fclose (file);
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }
    return f;
}


double compute_f_and_gradient_f (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    double counter [MAXNUMPARAMS];    
    double logZ_gradient;
    double f;
    f = 0;
    int real_str_found;
    double numerator [MAXNUMPARAMS];
    
    double denominator;      
    char structure[MAXSLEN];
    double enthalpy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];    
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    
    int seen = 0;
    while (!feof (file))
    {
        seen++;
        // fixed bug: real_str_found was initialized with 0 outside of the while loop
        real_str_found = 0;
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;    
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter, 1);

        // now fold sequence
        s_min_folding *min_fold = new s_min_folding (sequence, restricted);
        min_energy = min_fold->s_simfold_restricted ();
        min_fold->return_structure (structure);
        delete min_fold;      
        
        s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
        sub_fold->set_limit (MAXSUBSTR);
        sub_fold->s_simfold_restricted (enthalpy);
        actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
        max_energy = tmp_energies[actual_num_str-1];
        delete sub_fold;
        
        for (i=0; i < num_params; i++)
        {
            numerator[i] = 0;
        }    
            
        // first compute the denominator, which is the same for all parameters
        denominator = 0;
        for (k=0; k < actual_num_str; k++)
        {
            // check if the real structure is one of the suboptimal ones
            if (strcmp (tmp_structures[k], real_structure) == 0)
                real_str_found = 1;
            // recompute the free energy, i.e. with the correct dangling ends
            energy = free_energy_simfold_restricted (sequence, tmp_structures[k], restricted);        
            denominator += exp ((-1) * energy * beta);
            count_each_structure_type (sequence, tmp_structures[k], counter, 1);
            for (i=0; i < num_params; i++)
            {
                numerator[i] += counter[i] * exp ((-1) * energy * beta);
            }
        }
        count_each_structure_type (sequence, real_structure, counter, 1);
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);        
        if (!real_str_found)
        {
            // if not found, add it
            denominator += exp ((-1) * energy * beta);                    
            for (i=0; i < num_params; i++)
            {
                numerator[i] += counter[i] * exp ((-1) * energy * beta);
            }    
        }
        // now the denominator and nominator are computed
        
        f += beta*energy + log(denominator);
        //printf ("logZ(%d) = %e\n", seen, log(denominator));
        for (i=0; i < num_params; i++)
        {
            logZ_gradient = numerator[i] / denominator;
            f_gradient[i] += counter[i] - logZ_gradient;
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], numerator[7621]/denominator, f_gradient[7621]);
        //printf ("logZ_grad[i] = %e\n", numerator[7621]/denominator);
    }      
    fclose (file);
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }    
    return f;
}


void compute_counts_vector_LP (char *input_file, double *total_counter)
// PRE:  the parameters have been read
// POST: given p input sequence/structure pairs, compute sum_{j=1}^p (c_j, cbar_j)
//          where c_j is the counts for the known structures
//          and cbar_j is the counts for the predicted mfe structure, with the given set of parameters
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char pred_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    double counter [MAXNUMPARAMS];    
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    
    for (i = 0; i < num_params; i++)
    {    
        total_counter[i] = 0;
    }    
    
    
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;    
        //printf ("===\nReal:\nSeq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter, 1);
        for (i = 0; i < num_params; i++)  total_counter[i] += counter[i];
        // now predict the structure        
        simfold_restricted (sequence, restricted, pred_structure);
        //printf ("===\nPred:\nSeq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, pred_structure, restricted);
        count_each_structure_type (sequence, pred_structure, counter, 1);
        for (i = 0; i < num_params; i++)  total_counter[i] -= counter[i];        
    }      
}

int compute_counts_matrix_LP_helper (FILE *file)
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char pred_structure[MAXSLEN];
    char restricted[MAXSLEN];
    double counter_real [MAXNUMPARAMS];    
    double counter_pred [MAXNUMPARAMS];    
    
    if (!get_info_from_file (file, sequence, real_structure, restricted)) 
        return 0;    
    //printf ("===\nReal:\nSeq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
    count_each_structure_type (sequence, real_structure, counter_real, 1);
    // now predict the structure        
    simfold_restricted (sequence, restricted, pred_structure);
    //printf ("===\nPred:\nSeq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, pred_structure, restricted);
    count_each_structure_type (sequence, pred_structure, counter_pred, 1);
    // write the counters to the std output
    printf ("[");
    for (int i = 0; i < num_params-1; i++)
    {
        printf ("%.2lf, ", counter_real[i] - counter_pred[i]);
    }
    printf ("%.2lf]", counter_real[num_params-1] - counter_pred[num_params-1]);  
    return 1;
}

void compute_counts_matrix_LP (char *input_file, int train_samples)
// PRE:  the parameters have been read
//       train_samples = # of training instances. -1 if all.
// POST: given p input sequence/structure pairs, compute and display the counts c_j - cbar_j
//          where c_j is the counts for the known structures
//          and cbar_j is the counts for the predicted mfe structure, with the given set of parameters
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char pred_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    double counter_real [MAXNUMPARAMS];    
    double counter_pred [MAXNUMPARAMS];    
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    
    int k;
    k = 0;

    printf ("[");
    compute_counts_matrix_LP_helper (file);    
    if (train_samples > 0)
    {
        for (i = 1; i < train_samples; i++)
        {
            printf (",\n ");
            if (! compute_counts_matrix_LP_helper (file))
            {
                printf ("Error in file\n");
                exit(1);
            }
        }
    }            
    else 
    {
        while (!feof (file))
        {
            printf (",\n ");
            if (! compute_counts_matrix_LP_helper (file))
            {
                printf ("Error in file\n");
                exit(1);
            }
        }        
    }      
    printf ("]\n");
    fclose (file);
}


void find_indeces_of_bbtypes (int &first, int &last, char *bbtype, int num_params)
// PRE: the string_params are filled
// assumes the bb of type are consecutive
{
  int i;
  int met_first = 0;
  last = -1;
  for (i=0; i < num_params; i++)
    {
      if (strstr (string_params[i], bbtype) != NULL)
        {
          if (!met_first)
            {
              met_first = 1;
              first = i;
            }
        }
      else
        {
          if (met_first)  // this must be the next after last
            {
              last = i-1;
              break;
            }
        }
    }        
  if (last == -1)
    last = num_params-1;
}


int choose_bbtype_randomly (int first, int last)
// PRE: the string_params are filled, the indeces first and last have been found
{
  int random_number;
  // generate a number between first and last
  random_number = first + (int) ((last-first+1.0)*rand()/(RAND_MAX+1.0));
  return random_number;
}



int generate_structure_withbb (char *sequence, char *known_structure, char *given_restricted, char *turner_structure, char structures[][MAXSLEN], double *old_counts, int threshold)
// written by Mirela in Sep, 2005
// PRE: the bbseq_left, bbstr_left, bbseq_right and bbstr_right are filled
//        i.e. call num_params = create_building_block_strings();
// First try to find if the building block bb_index can exist in any structure.
//   If found, restrict that part, and call simfold_restricted
//   Return the number of structures found
//   Look for at most threshold structures
//   num_params = total number of parameters
//   true_fe = the estimated free energy of the true structure. 
//        We consider only those structures whose energies are <= true_fe
//   Update old_counts every time I consider a structure 
{
    char *position_left, *position_right;
    char *position_left2, *position_right2;  // for tstacki
    char restricted[MAXSLEN];
    int seqlen, i, numstr;   
    int bb_index;
    double counter_other[MAXNUMPARAMS];
    double energies[MAXSUBSTR];
    double known_fe;
    int considered;
    
    seqlen = strlen(sequence);
    numstr = 0;
    bb_index = 0;
           
    if (given_restricted[0] == '\0')
        known_fe = free_energy_simfold (sequence, known_structure);
    else    
        known_fe = free_energy_simfold_restricted (sequence, known_structure, given_restricted);
    //printf ("Known fe: %.2lf\n", known_fe);
    
    // first find bb_index, which is < threshold
    while (bb_index < num_params)
    {
        position_left = strstr (sequence, bbseq_left[bb_index]);            
        if ((old_counts[bb_index] >= threshold) ||
            (strcmp (bbseq_left[bb_index], "") == 0) ||     // building block not implemented
            (strstr (string_params[bb_index], "dangle_") != NULL))    // not implemented yet
        {
            bb_index++;            
            continue;  
        }            
        //printf ("Looking for parameter %s\n", string_params[bb_index]);        
        considered = 0;
        while (position_left != NULL && !considered)    // include only one structure having some building block, no more
        //while (position_left != NULL && old_counts[bb_index] < threshold)
        {
            position_right = position_left + strlen (bbseq_left[bb_index])+TURN;  //?
            // check I'm not out of bounds
            if (position_right - sequence >= seqlen)
                break;
            position_right = strstr (position_right, bbseq_right[bb_index]);
            while (position_right != NULL && old_counts[bb_index] < threshold)
            {
                // found a new place for my building block
                for (i=0; i < seqlen; i++)   restricted[i] = '_'; restricted [seqlen] = '\0'; 
                replace_str_piece (restricted, position_left-sequence, bbstr_left[bb_index]);
                replace_str_piece (restricted, position_right-sequence, bbstr_right[bb_index]);
                // if it if type tstackh, then everything in between bbstr_left and bbstr_right must be "...."
                if (strstr (string_params[bb_index], "tstackh") != NULL)
                {
                    if (position_right - position_left > MAXLOOP)
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    for (i=0; i < position_right - position_left - strlen (bbstr_left[bb_index]); i++)
                        restricted[position_left - sequence + strlen (bbstr_left[bb_index]) + i] = '.';
                }
                if (strstr (string_params[bb_index], "tstacki") != NULL)
                {
                    position_left2 = NULL;
                    position_right2 = NULL;
                    int found;
                    found = 0;
                    int bb_index2;
                    int tried = 0;
                    // we have to find another tstacki inside, at the right distance
                    // the second stacki can be any stacki, not only the current one
                    int first, last, k;
                    find_indeces_of_bbtypes (first, last, "tstacki", num_params);
                    // try about 20 times
                    for (k=0; k < 20; k++)
                    {
                        bb_index2 = choose_bbtype_randomly (first, last);
                        //printf ("first=%d, last=%d, chosen = %d\n", first, last, bb_index2);
                        if (strstr (string_params[bb_index2], "tstacki") == NULL)              
                        {
                            printf ("bbtype was not chosen correctly\n");
                            exit(1);
                        }
                        //printf ("The second tstacki chosen was: %s\n", string_params[bb_index2]);
                        tried = 1;
                        //printf ("Looking for %s\n",  bbseq_right[bb_index2]);
                        position_left2 = strstr (position_left + strlen (bbseq_left[bb_index]), bbseq_right[bb_index2]);
                        while (position_left2 != NULL)
                        {
                            if (position_left2 > position_right -  strlen(bbseq_right[bb_index2]))   // is not inside
                                break;
                            position_right2 = strstr (position_left2 + strlen (bbseq_right[bb_index]) + TURN, bbseq_left[bb_index2]);
                            while (position_right2 != NULL)
                            {
                                if (position_right2 > position_right - strlen(bbseq_left[bb_index2]))
                                    break;
                                // now we should have both 1 and 2 somewhere
                                // now make sure the distance is right
                                if ((position_left2 - position_left == 2 && position_right - position_right2 == 2) || 
                                    (position_left2 - position_left + position_right - position_right2 > MAXLOOP))
                                    // not good, they shouldn't be 2 and 2 or longer than MAXLOOP
                                    // keep trying
                                {
                                    position_right2 = strstr (position_right2+1, bbseq_left[bb_index2]);
                                    continue;
                                }
                                // finally, found:
                                restricted[position_left2-sequence] = '.';
                                restricted[position_left2-sequence+1] = '(';
                                restricted[position_right2-sequence] = ')';
                                restricted[position_right2-sequence+1] = '.';
                                // now fill everything in between
                                for (i=position_left - sequence + strlen (bbseq_left[bb_index]); i < position_left2-sequence; i++)
                                    restricted[i] = '.';
                                for (i=position_right2 - sequence + strlen (bbseq_right[bb_index2]); i < position_right-sequence; i++)
                                    restricted[i] = '.';                      
                                found = 1;
                                break;
                            }  // end while position_right2
                            if (found) break;
                            position_left2 = strstr (position_left2+1, bbseq_right[bb_index2]);
                        }  // end while position_right1
                        if (found) break;
                    }  // end for bb_index2
                    if (!found) 
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    //printf ("\t%s\n", restricted);
        
                } // end if tstacki
                //printf ("\t%s\n", restricted);
                
                // now make sure restricted is compatible with given_restricted
                if (given_restricted[0] != '\0')
                {
                    //printf ("--%s\n++%s\n", given_restricted, restricted);
                    int comp = restricted_compatible (given_restricted, restricted);
                    if (comp)
                    {
                        //printf ("**%s\n\n", restricted);
                    }
                    else
                    {
                        //printf ("INCOMPATIBLE!\n\n");
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);                        
                        continue;
                    }
                }                
                energies[numstr] = simfold_restricted (sequence, restricted, structures[numstr]);
                
                // test that this structure is not the same as the known_structure, turner_structure, or any other structure generated so far
                int to_consider = 1;
                if (strcmp (known_structure, structures[numstr]) == 0)          to_consider = 0;
                else if (strcmp (turner_structure, structures[numstr]) == 0)    to_consider = 0;
                else
                {
                    for (i=0; i < numstr; i++)
                        if (strcmp (structures[i], structures[numstr]) == 0)
                            to_consider = 0;
                }
                
                if (to_consider && energies[numstr] <= known_fe)    // only then consider it
                {                                        
                    // update old_counts, by adding the new counts
                    count_each_structure_type (sequence, structures[numstr], old_counts, 0);                               
                    numstr++;
                    considered = 1;
                    if (numstr >= MAXSUBSTR)     // I have to make it stop somewhere, just in case, otherwise it can run out of bounds
                        return numstr;
                }    
                position_right = strstr (position_right+1, bbseq_right[bb_index]);
            } // end while position_right
            position_left = strstr (position_left+1, bbseq_left[bb_index]);
        }    // end while position_left
        bb_index++;
    }    
    return numstr;

  // stack, int11, int21, int22 are good to go - done
  // tstackh: everything in between bbstr_left and bbstr_right must be "...." - done
  // tstacki: in between left and right, there must be another tstacki, and in between, we must have "..." "..." of appropriate length - done
  // dangle_top and dangle_bot - must be in exterior loop (too complicated in multi loop) - NOT DONE YET
}



/*
int generate_structure_withoutbb (char *sequence, char *known_structure, char *given_restricted, char *turner_structure, char structures[][MAXSLEN], int *old_counts, int threshold)
// written by Mirela in Jan, 2006
// If known_structure has a "critical" building block, i.e. which does not appear in the thermodynamic set, then predict a few structures which don't have this building block
{
    char *position_left, *position_right;
    char *position_left2, *position_right2;  // for tstacki
    char restricted[MAXSLEN];
    int seqlen, i, numstr;   
    int bb_index;
    int counter_known[MAXNUMPARAMS];
    int counter_other[MAXNUMPARAMS];
    double energies[MAXSUBSTR];
    
    double known_fe;
    int considered;
    
    // first see which of the critical building blocks are in the known structure
    count_each_structure_type (sequence, structure_pred, counter_known, 1);
    for (i=0; i < numparams; i++)
    {
        if (counter_known[i] > 0 && old_counts[i] < threshold)
        {
            position_left = strstr (sequence, bbseq_left[bb_index]);         
        }
    }
    
    
    
      
    
    seqlen = strlen(sequence);
    numstr = 0;
    bb_index = 0;
           
    if (given_restricted[0] == '\0')
        known_fe = free_energy_simfold (sequence, known_structure);
    else    
        known_fe = free_energy_simfold_restricted (sequence, known_structure, given_restricted);
    //printf ("Known fe: %.2lf\n", known_fe);
    
    // first find bb_index, which is < threshold
    while (bb_index < num_params)
    {
        position_left = strstr (sequence, bbseq_left[bb_index]);            
        if ((old_counts[bb_index] >= threshold) ||
            (strcmp (bbseq_left[bb_index], "") == 0) ||     // building block not implemented
            (strstr (string_params[bb_index], "dangle_") != NULL))    // not implemented yet
        {
            bb_index++;            
            continue;  
        }            
        //printf ("Looking for parameter %s\n", string_params[bb_index]);        
        considered = 0;
        while (position_left != NULL && !considered)    // include only one structure having some building block, no more
        //while (position_left != NULL && old_counts[bb_index] < threshold)
        {
            position_right = position_left + strlen (bbseq_left[bb_index])+TURN;  //?
            // check I'm not out of bounds
            if (position_right - sequence >= seqlen)
                break;
            position_right = strstr (position_right, bbseq_right[bb_index]);
            while (position_right != NULL && old_counts[bb_index] < threshold)
            {
                // found a new place for my building block
                for (i=0; i < seqlen; i++)   restricted[i] = '_'; restricted [seqlen] = '\0'; 
                replace_str_piece (restricted, position_left-sequence, bbstr_left[bb_index]);
                replace_str_piece (restricted, position_right-sequence, bbstr_right[bb_index]);
                // if it if type tstackh, then everything in between bbstr_left and bbstr_right must be "...."
                if (strstr (string_params[bb_index], "tstackh") != NULL)
                {
                    if (position_right - position_left > MAXLOOP)
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    for (i=0; i < position_right - position_left - strlen (bbstr_left[bb_index]); i++)
                        restricted[position_left - sequence + strlen (bbstr_left[bb_index]) + i] = '.';
                }
                if (strstr (string_params[bb_index], "tstacki") != NULL)
                {
                    position_left2 = NULL;
                    position_right2 = NULL;
                    int found;
                    found = 0;
                    int bb_index2;
                    int tried = 0;
                    // we have to find another tstacki inside, at the right distance
                    // the second stacki can be any stacki, not only the current one
                    int first, last, k;
                    find_indeces_of_bbtypes (first, last, "tstacki", num_params);
                    // try about 20 times
                    for (k=0; k < 20; k++)
                    {
                        bb_index2 = choose_bbtype_randomly (first, last);
                        //printf ("first=%d, last=%d, chosen = %d\n", first, last, bb_index2);
                        if (strstr (string_params[bb_index2], "tstacki") == NULL)              
                        {
                            printf ("bbtype was not chosen correctly\n");
                            exit(1);
                        }
                        //printf ("The second tstacki chosen was: %s\n", string_params[bb_index2]);
                        tried = 1;
                        //printf ("Looking for %s\n",  bbseq_right[bb_index2]);
                        position_left2 = strstr (position_left + strlen (bbseq_left[bb_index]), bbseq_right[bb_index2]);
                        while (position_left2 != NULL)
                        {
                            if (position_left2 > position_right -  strlen(bbseq_right[bb_index2]))   // is not inside
                                break;
                            position_right2 = strstr (position_left2 + strlen (bbseq_right[bb_index]) + TURN, bbseq_left[bb_index2]);
                            while (position_right2 != NULL)
                            {
                                if (position_right2 > position_right - strlen(bbseq_left[bb_index2]))
                                    break;
                                // now we should have both 1 and 2 somewhere
                                // now make sure the distance is right
                                if ((position_left2 - position_left == 2 && position_right - position_right2 == 2) || 
                                    (position_left2 - position_left + position_right - position_right2 > MAXLOOP))
                                    // not good, they shouldn't be 2 and 2 or longer than MAXLOOP
                                    // keep trying
                                {
                                    position_right2 = strstr (position_right2+1, bbseq_left[bb_index2]);
                                    continue;
                                }
                                // finally, found:
                                restricted[position_left2-sequence] = '.';
                                restricted[position_left2-sequence+1] = '(';
                                restricted[position_right2-sequence] = ')';
                                restricted[position_right2-sequence+1] = '.';
                                // now fill everything in between
                                for (i=position_left - sequence + strlen (bbseq_left[bb_index]); i < position_left2-sequence; i++)
                                    restricted[i] = '.';
                                for (i=position_right2 - sequence + strlen (bbseq_right[bb_index2]); i < position_right-sequence; i++)
                                    restricted[i] = '.';                      
                                found = 1;
                                break;
                            }  // end while position_right2
                            if (found) break;
                            position_left2 = strstr (position_left2+1, bbseq_right[bb_index2]);
                        }  // end while position_right1
                        if (found) break;
                    }  // end for bb_index2
                    if (!found) 
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    //printf ("\t%s\n", restricted);
        
                } // end if tstacki
                //printf ("\t%s\n", restricted);
                
                // now make sure restricted is compatible with given_restricted
                if (given_restricted[0] != '\0')
                {
                    //printf ("--%s\n++%s\n", given_restricted, restricted);
                    int comp = restricted_compatible (given_restricted, restricted);
                    if (comp)
                    {
                        //printf ("**%s\n\n", restricted);
                    }
                    else
                    {
                        //printf ("INCOMPATIBLE!\n\n");
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);                        
                        continue;
                    }
                }                
                energies[numstr] = simfold_restricted (sequence, restricted, structures[numstr]);
                
                // test that this structure is not the same as the known_structure, turner_structure, or any other structure generated so far
                int to_consider = 1;
                if (strcmp (known_structure, structures[numstr]) == 0)          to_consider = 0;
                else if (strcmp (turner_structure, structures[numstr]) == 0)    to_consider = 0;
                else
                {
                    for (i=0; i < numstr; i++)
                        if (strcmp (structures[i], structures[numstr]) == 0)
                            to_consider = 0;
                }
                
                if (to_consider && energies[numstr] <= known_fe)    // only then consider it
                {                                        
                    // update old_counts, by adding the new counts
                    count_each_structure_type (sequence, structures[numstr], old_counts, 0);                               
                    numstr++;
                    considered = 1;
                    if (numstr >= MAXSUBSTR)     // I have to make it stop somewhere, just in case, otherwise it can run out of bounds
                        return numstr;
                }    
                position_right = strstr (position_right+1, bbseq_right[bb_index]);
            } // end while position_right
            position_left = strstr (position_left+1, bbseq_left[bb_index]);
        }    // end while position_left
        bb_index++;
    }    
    return numstr;

  // stack, int11, int21, int22 are good to go - done
  // tstackh: everything in between bbstr_left and bbstr_right must be "...." - done
  // tstacki: in between left and right, there must be another tstacki, and in between, we must have "..." "..." of appropriate length - done
  // dangle_top and dangle_bot - must be in exterior loop (too complicated in multi loop) - NOT DONE YET
}

*/

int restricted_compatible (char *given_restricted, char *restricted)
// assume given_restricted doesn't have dots
// return 1 if the two are compatible
// given_restricted is input variable
// restricted is input-output, and it will contain the joint restriction
{
    int given_p_table[MAXSLEN];
    int p_table[MAXSLEN];
    detect_original_pairs (given_restricted, given_p_table);
    detect_original_pairs (restricted, p_table);
    int len = strlen (restricted);
    int i;
    int compatible = 1;
    for (i=0; i < len; i++)
    {
        if (p_table[i] == -1)    // dot
        {
            if (given_p_table[i] >= 0)
            {
                compatible = 0;
                break;
            }    
        }    
        else if (p_table[i] >= 0)
        {
            if (given_p_table[i] >= 0 && given_p_table[i] != p_table[i])
            {
                compatible = 0;
                break;
            }            
        }
    }
    if (compatible)
    {
        for (i=0; i < len; i++)
        {
            if (given_p_table[i] >= -1)
                restricted[i] = given_restricted[i];
        }
    }
    
    return compatible;
}


int generate_structure_withbb_many_thresholds (char *sequence, char structures[][MAXSLEN], double energies[], double *old_counts, int many_thresholds[], double true_fe, int num_params)
// written by Mirela in Sep, 2005
// PRE: the bbseq_left, bbstr_left, bbseq_right and bbstr_right are filled
//        i.e. call num_params = create_building_block_strings();
// First try to find if the building block bb_index can exist in any structure.
//   If found, restrict that part, and call simfold_restricted
//   Return the number of structures found
//   Look for at most threshold structures
//   num_params = total number of parameters
//   true_fe = the estimated free energy of the true structure. 
//        We consider only those structures whose energies are <= true_fe
//   Update old_counts every time I consider a structure 

// many_thresholds is a threshold per parameter (INF of no threshols wanted)
{
    char *position_left, *position_right;
    char *position_left2, *position_right2;  // for tstacki
    char restricted[MAXSLEN];
    int seqlen, i, numstr;   
    int bb_index;
    double counter_other[MAXNUMPARAMS];
    
    seqlen = strlen(sequence);
    numstr = 0;
    bb_index = 0;
    
    // first find bb_index, which is < threshold
    while (bb_index < num_params)
    {        
        position_left = strstr (sequence, bbseq_left[bb_index]);            
        if ((many_thresholds[bb_index] >= INF) ||
            (old_counts[bb_index] >= many_thresholds[bb_index]) ||
            (strcmp (bbseq_left[bb_index], "") == 0) ||     // building block not implemented
            (strstr (string_params[bb_index], "dangle_") != NULL))    // not implemented yet
        {
            bb_index++;            
            continue;  
        }            
        while (position_left != NULL && old_counts[bb_index] < many_thresholds[bb_index])
        {
            position_right = position_left + strlen (bbseq_left[bb_index])+TURN;  //?
            // check I'm not out of bounds
            if (position_right - sequence >= seqlen)
                break;
            position_right = strstr (position_right, bbseq_right[bb_index]);
            while (position_right != NULL && old_counts[bb_index] < many_thresholds[bb_index])
            {
                // found a new place for my building block
                for (i=0; i < seqlen; i++)   restricted[i] = '_'; restricted [seqlen] = '\0'; 
                replace_str_piece (restricted, position_left-sequence, bbstr_left[bb_index]);
                replace_str_piece (restricted, position_right-sequence, bbstr_right[bb_index]);
                // if it if type tstackh, then everything in between bbstr_left and bbstr_right must be "...."
                if (strstr (string_params[bb_index], "tstackh") != NULL)
                {
                    if (position_right - position_left > MAXLOOP)
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    for (i=0; i < position_right - position_left - strlen (bbstr_left[bb_index]); i++)
                        restricted[position_left - sequence + strlen (bbstr_left[bb_index]) + i] = '.';
                }
                if (strstr (string_params[bb_index], "tstacki") != NULL)
                {
                    position_left2 = NULL;
                    position_right2 = NULL;
                    int found;
                    found = 0;
                    int bb_index2;
                    int tried = 0;
                    // we have to find another tstacki inside, at the right distance
                    // the second stacki can be any stacki, not only the current one
                    int first, last, k;
                    find_indeces_of_bbtypes (first, last, "tstacki", num_params);
                    // try about 20 times
                    for (k=0; k < 20; k++)
                    {
                        bb_index2 = choose_bbtype_randomly (first, last);
                        //printf ("first=%d, last=%d, chosen = %d\n", first, last, bb_index2);
                        if (strstr (string_params[bb_index2], "tstacki") == NULL)              
                        {
                            printf ("bbtype was not chosen correctly\n");
                            exit(1);
                        }
                        tried = 1;
                        position_left2 = strstr (position_left + strlen (bbseq_left[bb_index]), bbseq_right[bb_index2]);
                        while (position_left2 != NULL)
                        {
                            if (position_left2 > position_right -  strlen(bbseq_right[bb_index2]))   // is not inside
                                break;
                            position_right2 = strstr (position_left2 + strlen (bbseq_right[bb_index]) + TURN, bbseq_left[bb_index2]);
                            while (position_right2 != NULL)
                            {
                                if (position_right2 > position_right - strlen(bbseq_left[bb_index2]))
                                    break;
                                // now we should have both 1 and 2 somewhere
                                // now make sure the distance is right
                                if ((position_left2 - position_left == 2 && position_right - position_right2 == 2) || 
                                    (position_left2 - position_left + position_right - position_right2 > MAXLOOP))
                                    // not good, they shouldn't be 2 and 2 or longer than MAXLOOP
                                    // keep trying
                                {
                                    position_right2 = strstr (position_right2+1, bbseq_left[bb_index2]);
                                    continue;
                                }
                                // finally, found:
                                restricted[position_left2-sequence] = '.';
                                restricted[position_left2-sequence+1] = '(';
                                restricted[position_right2-sequence] = ')';
                                restricted[position_right2-sequence+1] = '.';
                                // now fill everything in between
                                for (i=position_left - sequence + strlen (bbseq_left[bb_index]); i < position_left2-sequence; i++)
                                    restricted[i] = '.';
                                for (i=position_right2 - sequence + strlen (bbseq_right[bb_index2]); i < position_right-sequence; i++)
                                    restricted[i] = '.';                      
                                found = 1;
                                break;
                            }  // end while position_right2
                            if (found) break;
                            position_left2 = strstr (position_left2+1, bbseq_right[bb_index]);
                        }  // end while position_right1
                        if (found) break;
                    }  // end for bb_index2
                    if (!found) 
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    //printf ("\t%s\n", restricted);
        
                } // end if tstacki
                //printf ("\t%s\n", restricted);
                energies[numstr] = simfold_restricted (sequence, restricted, structures[numstr]);
                if (energies[numstr] <= true_fe)    // only then consider it
                {                    
                    // update old_counts, by adding the new counts
                    count_each_structure_type (sequence, structures[numstr], old_counts, 0);                               
                    numstr++;
                    if (numstr >= MAXSUBSTR)     // I have to make it stop somewhere, just in case, otherwise it can run out of bounds
                        return numstr;
                }    
                position_right = strstr (position_right+1, bbseq_right[bb_index]);
            } // end while position_right
            position_left = strstr (position_left+1, bbseq_left[bb_index]);
        }    // end while position_left
        bb_index++;
    }    
    return numstr;

  // stack, int11, int21, int22 are good to go - done
  // tstackh: everything in between bbstr_left and bbstr_right must be "...." - done
  // tstacki: in between left and right, there must be another tstacki, and in between, we must have "..." "..." of appropriate length - done
  // dangle_top and dangle_bot - must be in exterior loop (too complicated in multi loop) - NOT DONE YET
}




void search_bb (char *sequence, double *old_counts, int threshold, int num_params)
// written by Mirela in Sep, 2005
// PRE: the bbseq_left, bbstr_left, bbseq_right and bbstr_right are filled
//        i.e. call num_params = create_building_block_strings();
// First try to find if the building block bb_index can exist in any structure.
//   If found, restrict that part, and call simfold_restricted
//   Return the number of structures found
//   Look for at most threshold structures
//   num_params = total number of parameters
//   true_fe = the estimated free energy of the true structure. 
//        We consider only those structures whose energies are <= true_fe
//   Update old_counts every time I consider a structure 
{
    char *position_left, *position_right;
    char *position_left2, *position_right2;  // for tstacki
    char restricted[MAXSLEN];
    int seqlen, i, numstr;   
    int bb_index;
    double counter_other[MAXNUMPARAMS];
    
    seqlen = strlen(sequence);
    numstr = 0;
    bb_index = 0;
    
    // first find bb_index, which is < threshold    
    while (bb_index < num_params)
    {
        position_left = strstr (sequence, bbseq_left[bb_index]);        
        if ((strcmp (bbseq_left[bb_index], "") == 0) ||     // building block not implemented
            (strstr (string_params[bb_index], "dangle_") != NULL))    // not implemented yet
        {
            bb_index++;
            continue;  
        }            
        while (position_left != NULL)
        {
            position_right = position_left + strlen (bbseq_left[bb_index])+TURN;  //?
            // check I'm not out of bounds
            if (position_right - sequence >= seqlen)
                break;
            position_right = strstr (position_right, bbseq_right[bb_index]);
            while (position_right != NULL)
            {
                // found a new place for my building block
                for (i=0; i < seqlen; i++)   restricted[i] = '_'; restricted [seqlen] = '\0'; 
                replace_str_piece (restricted, position_left-sequence, bbstr_left[bb_index]);
                replace_str_piece (restricted, position_right-sequence, bbstr_right[bb_index]);
                // if it if type tstackh, then everything in between bbstr_left and bbstr_right must be "...."
                if (strstr (string_params[bb_index], "tstackh") != NULL)
                {
                    if (position_right - position_left > MAXLOOP)
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    for (i=0; i < position_right - position_left - strlen (bbstr_left[bb_index]); i++)
                        restricted[position_left - sequence + strlen (bbstr_left[bb_index]) + i] = '.';
                }
                if (strstr (string_params[bb_index], "tstacki") != NULL)
                {
                    position_left2 = NULL;
                    position_right2 = NULL;
                    int found;
                    found = 0;
                    int bb_index2;
                    int tried = 0;
                    // we have to find another tstacki inside, at the right distance
                    // the second stacki can be any stacki, not only the current one
                    int first, last, k;
                    find_indeces_of_bbtypes (first, last, "tstacki", num_params);
                    // try about 20 times
                    for (k=0; k < 20; k++)
                    {
                        bb_index2 = choose_bbtype_randomly (first, last);
                        //printf ("first=%d, last=%d, chosen = %d\n", first, last, bb_index2);
                        if (strstr (string_params[bb_index2], "tstacki") == NULL)              
                        {
                            printf ("bbtype was not chosen correctly\n");
                            exit(1);
                        }
                        tried = 1;
                        position_left2 = strstr (position_left + strlen (bbseq_left[bb_index]), bbseq_right[bb_index2]);
                        while (position_left2 != NULL)
                        {
                            if (position_left2 > position_right -  strlen(bbseq_right[bb_index2]))   // is not inside
                                break;
                            position_right2 = strstr (position_left2 + strlen (bbseq_right[bb_index]) + TURN, bbseq_left[bb_index2]);
                            while (position_right2 != NULL)
                            {
                                if (position_right2 > position_right - strlen(bbseq_left[bb_index2]))
                                    break;
                                // now we should have both 1 and 2 somewhere
                                // now make sure the distance is right
                                if ((position_left2 - position_left == 2 && position_right - position_right2 == 2) || 
                                    (position_left2 - position_left + position_right - position_right2 > MAXLOOP))
                                    // not good, they shouldn't be 2 and 2 or longer than MAXLOOP
                                    // keep trying
                                {
                                    position_right2 = strstr (position_right2+1, bbseq_left[bb_index2]);
                                    continue;
                                }
                                // finally, found:
                                restricted[position_left2-sequence] = '.';
                                restricted[position_left2-sequence+1] = '(';
                                restricted[position_right2-sequence] = ')';
                                restricted[position_right2-sequence+1] = '.';
                                // now fill everything in between
                                for (i=position_left - sequence + strlen (bbseq_left[bb_index]); i < position_left2-sequence; i++)
                                    restricted[i] = '.';
                                for (i=position_right2 - sequence + strlen (bbseq_right[bb_index2]); i < position_right-sequence; i++)
                                    restricted[i] = '.';                      
                                found = 1;
                                break;
                            }  // end while position_right2
                            if (found) break;
                            position_left2 = strstr (position_left2+1, bbseq_right[bb_index]);
                        }  // end while position_right1
                        if (found) break;
                    }  // end for bb_index2
                    if (!found) 
                    {
                        position_right = strstr (position_right+1, bbseq_right[bb_index]);
                        continue;
                    }
                    //printf ("\t%s\n", restricted);
        
                } // end if tstacki
                old_counts[bb_index]++;
                position_right = strstr (position_right+1, bbseq_right[bb_index]);
            } // end while position_right
            position_left = strstr (position_left+1, bbseq_left[bb_index]);
        }    // end while position_left
        bb_index++;
    }    

  // stack, int11, int21, int22 are good to go - done
  // tstackh: everything in between bbstr_left and bbstr_right must be "...." - done
  // tstacki: in between left and right, there must be another tstacki, and in between, we must have "..." "..." of appropriate length - done
  // dangle_top and dangle_bot - must be in exterior loop (too complicated in multi loop) - NOT DONE YET
}




void fill_data_structures_with_new_parameters_fixed_dangles (char *filename, char *dangfilename)
// reads all params from filename, except the dangling parameters, which reads from a different file
// assumes the fm363 model  

  // Mirela: Ian 11, 2007
  // reads parameters from a file, and writes them in the internal data structures
  // PRE: first read the actual standard parameters, to be able to figure out which of them are
  // < INF, and maybe to also keep some old values.
{
    int index;
    int i, j, k, l, m, n, o, p;
    FILE *file;
    char buffer[100];
    double param;
    int line = 0;

    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)
                        for (n=0; n < NUCL; n++)
                        {
                            int11[i][j][k][l][m][n] = INF;
                        }

    
    //printf ("FILENAME: %s\n", filename);
    if ((file = fopen (filename, "r")) == NULL)
    {
        giveup ("Cannot open file", filename);
    }
    
    index = 0;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (stack[i][j][k][l] < INF)
                {
                    // exclude duplicates
                    // stack[i][j][k][l] is the same as stack[l][k][j][i]
                    if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                    {
                        fgets (buffer, sizeof(buffer), file);
                        line++;
                        sscanf (buffer, "%lf\n", &param);
                        //printf ("\t%lf\n", param);
                        // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                        {
                            param *= 100;
                            stack[i][j][k][l] = (int)(round(param));
                            // add the duplicate too
                            stack[l][k][j][i] = (int)(round(param));
                        }
                        //sprintf (string_params[index++], "stack[%d][%d][%d][%d]", i, j, k, l);
                    }
                }
            }
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (tstackh[i][j][k][l] < INF)
                {
                    // no duplicates here
                    fgets (buffer, sizeof(buffer), file);
                    line++;
                    sscanf (buffer, "%lf", &param);
                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                        param *= 100;
                        tstackh[i][j][k][l] = (int)(round(param));
                    }
                    //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
                }
            }
    // use only 3 parameters, and fill the tstacki data structure by applying the Mathews 99 rules
    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_AU_closure = (int)(round(param));
    
    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_AG_mismatch = (int)(round(param));

    fgets (buffer, sizeof(buffer), file);
    line++;
    sscanf (buffer, "%lf", &param);
    param *= 100;
    misc.internal_UU_mismatch = (int)(round(param));
    
    // fill the tstacki data structure a bit later, after we read AU_penalty
    
    /*
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
            {
                if (tstacki[i][j][k][l] < INF)
                {
                    // no duplicates here
                    fgets (buffer, sizeof(buffer), file);
                    sscanf (buffer, "%lf", &param);
                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                        param *= 100;
                        tstacki[i][j][k][l] = (int)(round(param));
                    }
                    //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
                }
            }
    */            
    if (!simple_internal_energy)
    {
        // do the few int 11 params: only those enclosed by CG and CG (any order), + 2 more params
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                        for (m=0; m < NUCL; m++)
                            for (n=0; n < NUCL; n++)
                            {
                                if ( (((i==C && j==G) || (i==G && j==C)) &&
                                     ((m==C && n==G) || (m==G && n==C)) &&
                                     !can_pair(k,l)) ||
                                     (watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                {
                                    //if (int11[i][j][k][l][m][n] < INF)
                                    {
                                        // exclude duplicates
                                        // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                                        if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                                        {
                                            fgets (buffer, sizeof(buffer), file);
                                            line++;
                                            sscanf (buffer, "%lf", &param);
                                            param *= 100;
                                            int11[i][j][k][l][m][n] = (int)(round(param));
                                            // add the duplicate too
                                            int11[n][m][l][k][j][i] = (int)(round(param));
                                        }
                                    }
                                }
                            }
        fgets (buffer, sizeof(buffer), file);
        line++;
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal11_basic_mismatch = (int)(round(param));
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal11_GG_mismatch = (int)(round(param));


        // fill the int11 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        if ( ((i==C && j==G) || (i==G && j==C)) &&
                                             ((m==C && n==G) || (m==G && n==C)))
                                        {
                                            if (can_pair(k,l))
                                            {
                                                int11[i][j][k][l][m][n] = misc.internal11_basic_mismatch;
                                                // add the duplicate too
                                                //int11[n][m][l][k][j][i] = misc.internal11_basic_mismatch;
                                            }
                                        }
                                        
                                        else
                                        {
                                            if (!(watson_crick(i,j) && watson_crick(m,n) && k==U && l==U))
                                            {
                                                if (k==G && l==G)
                                                {
                                                    int11[i][j][k][l][m][n] = misc.internal11_GG_mismatch;
                                                    // add the duplicate too
                                                    //int11[n][m][l][k][j][i] = misc.internal11_GG_mismatch;
                                                }
                                                else
                                                    int11[i][j][k][l][m][n] = misc.internal11_basic_mismatch;
                                                if (has_AU_penalty(i,j))
                                                    int11[i][j][k][l][m][n] += misc.internal_AU_closure;
                                                if (has_AU_penalty(m,n))
                                                    int11[i][j][k][l][m][n] += misc.internal_AU_closure;
                                            }
                                        }
                                        
                                        // round it to match Turner parameters
                                        //if (int11[i][j][k][l][m][n] % 10 == 5) int11[i][j][k][l][m][n] += 5;
                                        //if (int11[i][j][k][l][m][n] % 10 == -5) int11[i][j][k][l][m][n] += 5;
                                    }
                }                                

        /*    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    {
                        if (int11[i][j][k][l][m][n] < INF)
                        {
                            // exclude duplicates
                            // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                            if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int11[i][j][k][l][m][n] = (int)(round(param));
                                    // add the duplicate too
                                    int11[n][m][l][k][j][i] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                            }
                        }
                    }
        */
                            
        // go with few int21 parameters, as in Mathews et al 1999
        // closed by CG
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            param *= 100;
                            int21[i][j][k][l][m][n][o] = (int)(round(param));
                        }
        // closed by GC                        
        i=G; j=C; m=G; n=C;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    if (!can_pair(k,l) && !can_pair(k,o))
                        if (int21[i][j][k][l][m][n][o] < INF)
                        {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            param *= 100;
                            int21[i][j][k][l][m][n][o] = (int)(round(param));
                        }
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal21_match = (int)(round(param));
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal21_AU_closure = (int)(round(param));
       
        // fill the int21 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        for(o=0; o < NUCL; o++)
                                        {
                                            if ((i==C && j==G && m==C && n==G) ||  // these are already filled above, except what can pair inside
                                                (i==G && j==C && m==G && n==C))
                                            {
                                                if (can_pair(k,l) || can_pair(k,o))
                                                    int21[i][j][k][l][m][n][o] = misc.internal21_match;
                                            }
                                            else
                                            {
                                                if (can_pair(k,l) || can_pair(k,o))
                                                    int21[i][j][k][l][m][n][o] = misc.internal21_match;
                                                // I approximate it to int, so that we obtain the same as by counte_each_structure_type and summing up
                                                else
                                                    int21[i][j][k][l][m][n][o] = (int)(int21[C][G][k][l][C][G][o]/2) +
                                                                                 (int)(int21[G][C][k][l][G][C][o]/2);
                                                if (has_AU_penalty(i,j))
                                                    int21[i][j][k][l][m][n][o] += misc.internal21_AU_closure;
                                                if (has_AU_penalty(m,n))    
                                                    int21[i][j][k][l][m][n][o] += misc.internal21_AU_closure;
                                            }
                                            // round it to match Turner parameters - seems to be inconsistent
                                            //if (int21[i][j][k][l][m][n][o] % 10 == 5) int21[i][j][k][l][m][n][o] += 5;
                                        }
                                    }
                }                                
        

                    
        /*                    
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        {
                        if (int21[i][j][k][l][m][n][o] < INF)
                            {
                            // no duplicates here
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                param *= 100;
                                int21[i][j][k][l][m][n][o] = (int)(round(param));
                                }
                            //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                            }
                        }
        */                        
        // go with the 53 parameters, like in Mathews et al 1999
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                for (k=0; k < NUCL; k++)
                    for (l=0; l < NUCL; l++)
                    {
                        n = i;
                        m = j;
                        p = k;
                        o = l;
                        if (watson_crick(i,j) && !watson_crick(k,l))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                    // add the duplicate too
                                    int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        /*
        i=C; j=G; m=C; n=G;
        for (k=0; k < NUCL; k++)
            for (l=0; l < NUCL; l++)
                for (o=0; o < NUCL; o++)
                    for (p=0; p < NUCL; p++)
                    {
                        if (!watson_crick(k,l) && !watson_crick(o,p))
                        {
                            // exclude duplicates
                            // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                            if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <=
                                n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                            {
                                fgets (buffer, sizeof(buffer), file);
                                sscanf (buffer, "%lf", &param);
                                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                {
                                    param *= 100;
                                    int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                    // add the duplicate too
                                    int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                }
                                //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                            }
                        }
                    }
        */            
        // then add the 4 deltas
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_same_size = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_different_size = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_1stable_1unstable = (int)(round(param));                
        
        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_delta_AC = (int)(round(param));

        fgets (buffer, sizeof(buffer), file);
        sscanf (buffer, "%lf", &param);
        param *= 100;
        misc.internal22_match = (int)(round(param));

        int ii, jj, mm, nn;
        // fill the int22 data structure
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
                if (can_pair(i,j))
                {
                    for (k=0; k < NUCL; k++)
                        for (l=0; l < NUCL; l++)
                            for (m=0; m < NUCL; m++)
                                for (n=0; n < NUCL; n++)
                                    if (can_pair(m,n))
                                    {
                                        for(o=0; o < NUCL; o++)
                                            for (p=0; p < NUCL; p++)
                                            {
                                                /*
                                                if (i==C && j==G && m==C && n==G)
                                                {
                                                    if(watson_crick(k,l) || watson_crick(o,p))
                                                    {
                                                        int22[i][j][k][l][m][n][o][p] = misc.internal22_match;
                                                    }
                                                    // else do nothing, it's parameter
                                                } */
                                                // if a closing pair is wobble, it's the same as if G would be A
                                                if (i==G && j==U)   ii = A;     else ii = i;
                                                if (i==U && j==G)   jj = A;     else jj = j;
                                                if (m==G && n==U)   mm = A;     else mm = m;
                                                if (m==U && n==G)   nn = A;     else nn = n;

                                                
                                                if (watson_crick(k,l) || watson_crick(o,p))
                                                {
                                                    int22[i][j][k][l][m][n][o][p] = misc.internal22_match;
                                                }
                                                else if (nn==ii && mm==jj && p==k && o==l)  // the UG closing pairs are the same as UA
                                                {
                                                    int22[i][j][k][l][m][n][o][p] = int22[ii][jj][k][l][mm][nn][o][p];
                                                }
                                                else //if (!(n==i && m==j && p==k && o==l))   // was already filled above
                                                {
                                                    int result = check_stability_and_size (k, l, o, p);
                                                    // I approximate it to int, so that we obtain the same as by counte_each_structure_type and summing up
                                                    int temp = (int)(int22[ii][jj][k][l][jj][ii][l][k]/2.0) +
                                                               (int)(int22[nn][mm][p][o][mm][nn][o][p]/2.0);
                                                    // rounf it to match Turner parameters
                                                    //if (temp%10 == 5) temp -= 5; if (temp%10 == -5) temp += 5;
                                                    int22[i][j][k][l][m][n][o][p] =  temp;
                                                    switch (result)
                                                    {
                                                        case 1: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_same_size; break;
                                                        case 2: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_different_size; break;
                                                        case 3: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_1stable_1unstable; break;
                                                        case 4: int22[i][j][k][l][m][n][o][p] += misc.internal22_delta_AC; break;
                                                        default: printf ("ERROR: result %d for k=%d, l=%d, o=%d, p=%d, ABORT!\n", result, k, l, o, p); exit(1);                                                
                                                    }                                                
                                                }                                        
                                            }
                                    }
                }                                
                        
        /*                        
        for (i=0; i < NUCL; i++)
            for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                for (m=0; m < NUCL; m++)
                    for (n=0; n < NUCL; n++)
                    for(o=0; o < NUCL; o++)
                        for (p=0; p < NUCL; p++)
                        {
                            if (int22[i][j][k][l][m][n][o][p] < INF)
                            {
                                // exclude duplicates
                                // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                                if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                                    n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                                {
                                    fgets (buffer, sizeof(buffer), file);
                                    sscanf (buffer, "%lf", &param);
                                    // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                                    {
                                        param *= 100;
                                        int22[i][j][k][l][m][n][o][p] = (int)(round(param));
                                        // add the duplicate too
                                        int22[n][m][p][o][j][i][l][k] = (int)(round(param));
                                    }
                                    //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                                }
                            }
                        }
        */                    
    }     // end if (!simple_internal_energy)

    // now get the dangling ends from a different file
    // **********************************************************************************************************
    FILE *dangfile;
    if ((dangfile = fopen (dangfilename, "r")) == NULL)
    {
        giveup ("Cannot open file", dangfilename);
    }
                                    
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
            {
            if (dangle_top[i][j][k] < INF)
                {
                // no duplicates here
                fgets (buffer, sizeof(buffer), dangfile);
                sscanf (buffer, "%lf", &param);
                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                    param *= 100;
                    dangle_top[i][j][k] = (int)(round(param));
                    }
                //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
                }
            }
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
        for (k=0; k < NUCL; k++)
            {
            if (dangle_bot[i][j][k] < INF)
                {
                // no duplicates here
                fgets (buffer, sizeof(buffer), dangfile);
                sscanf (buffer, "%lf", &param);
                // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                    {
                    param *= 100;
                    dangle_bot[i][j][k] = (int)(round(param));
                    }
                //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
                }
            }
    fclose(dangfile);
    // **********************************************************************************************************            

            
    int start;        
    if (!simple_internal_energy)
        start = 4;
    else
        start = 1;                
    for (i=start; i <= MAXLOOP_I; i++)
        {
        if (internal_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                internal_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
            }
        }
    for (i=1; i <= MAXLOOP_B; i++)
        {
        if (bulge_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);    
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                bulge_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
            }
        }
    for (i=1; i <= MAXLOOP_H; i++)
        {
        if (hairpin_penalty_by_size[i] < INF)
            {
            // no duplicates here
            fgets (buffer, sizeof(buffer), file);
            sscanf (buffer, "%lf", &param);
            // if (param != 0)    // put Turner's parameters if it's 0, but I can do this in the learn.pl file
                {
                param *= 100;
                hairpin_penalty_by_size[i] = (int)(round(param));
                }
            //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
            }
        }
    //fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
    //    misc.param_greater30 = (int)(round(param));
    
    // set a fixed value to param_greater_30 for now
    misc.param_greater30 = 1.079;
    
    //sprintf (string_params[index++], "misc.param_greater30");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.terminal_AU_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.terminal_AU_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_GGG = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_GGG");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c1 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c1");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c2 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c2");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.hairpin_c3 = (int)(round(param));
    //sprintf (string_params[index++], "misc.hairpin_c3");

    // fill the tstacki data structure, now that we also have terminal_AU_penalty - actually I removed it from here
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (!can_pair (i, j))
                        tstacki[i][j][k][l] = INF;
                    else    
                    {
                        tstacki[i][j][k][l] = 0;
                        if (((i == A || i == G) && j == U) ||
                            ((j == A || j == G) && i == U))
                        {
                            tstacki[i][j][k][l] += misc.internal_AU_closure;
                        }
                        if ((k == A && l == G) ||
                            (l == A && k == G))
                        {
                            tstacki[i][j][k][l] += misc.internal_AG_mismatch;
                        }
                        if (k == U && l == U)
                        {
                            tstacki[i][j][k][l] += misc.internal_UU_mismatch;
                        }
                    }
                }

    
    // TODO
    // keep them fixed for now    
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); 
    // sprintf (string_params[index++], "misc.asymmetry_penalty_max_correction");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[0]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[1]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
    // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
    // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.
    
    // to come back!!!
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
    //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
    //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");
    
    //fgets (buffer, sizeof(buffer), file);
    //sprintf (string_params[index++], "misc.gail_rule");
    // keep this fixed 
    misc.gail_rule = 1;
    
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_offset = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_offset");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_helix_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_helix_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.multi_free_base_penalty = (int)(round(param));
    //sprintf (string_params[index++], "misc.multi_free_base_penalty");
    fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
    // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
        misc.intermolecular_initiation = (int)(round(param));
    //sprintf (string_params[index++], "misc.intermolecular_initiation");
    
    for(i=0; i < nb_triloops; i++)
        {
        fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
        // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
            triloop[i].energy = (int)(round(param));
        //sprintf (string_params[index++], "triloop[%d].energy", i);
        }
    
    for(i=0; i < nb_tloops; i++)
        {
        fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
        // if (param != 0)      // put Turner's parameters if it's 0, but I can do this in the learn.pl file
            tloop[i].energy = (int)(round(param));
        //sprintf (string_params[index++], "tloop[%d].energy", i);
        }
    
    fclose (file);
    //printf ("****** stack[1][2][0][3] = %d\n", stack[1][2][0][3]);
}


